import numpy as np
import scipy as sp
from numpy import ma

from pymake.util.math import lognormalize, categorical, sorted_perm, adj_to_degree, gem
from ml.model.modelbase import SVB



class immsb_scvb(SVB):

    # *args ?
    def _init_params(self):
        self.limit_elbo_diff = 1e-3
        ### The time Limitations are @heere
        # frontend integration ?
        _len = {}
        _len['K'] = self.expe.get('K')
        _len['N'] = self.frontend.getN()
        _len['nfeat'] = self.frontend.get_nfeat()
        data_ma = self.frontend.data_ma
        _len['nnz'] = self.frontend.ma_nnz()
        _len['nnz_t'] = self.frontend.ma_nnz_t()
        _len['dims'] = self.frontend.ma_dims()
        _len['ones'] = (data_ma == 1).sum()
        _len['zeros'] = (data_ma == 0).sum()
        self._len = _len

        self.iterations = self.expe.get('iterations', 1)

        # Chunk Parameters
        self.chunk_size = self._get_chunk()

        self.chunk_len = self._len['nnz'] / self.chunk_size

        if self.chunk_len < 1:
            self.chunk_size = self._len['nnz']
            self.chunk_len = 1

        self._init_gradient()
        #self._init_gradienta()

        # Hyperparams
        delta = self.expe['hyperparams']['delta']
        self.hyper_phi = np.asarray(delta) if isinstance(delta, (np.ndarray, list, tuple)) else np.asarray([delta] * self._len['nfeat'])
        self.hyper_theta = np.asarray([1.0 / (i + np.sqrt(self._len['K'])) for i in range(self._len['K'])])
        self.hyper_theta /= self.hyper_theta.sum()

        # Sufficient Statistics
        #self._ss = self._random_s_init()
        self._ss = self._random_ss_init()
        #self._random_cvb_init()

        self.frontend._set_rawdata_for_likelihood_computation()

    def _init_gradient(self):
        self._timestep_a = 0
        self._timestep_b = 0
        self._chi_a = self.expe.get('chi_a', 5)
        self._tau_a = self.expe.get('tau_a', 10)
        self._kappa_a = self.expe.get('kappa_a', 0.9)
        self._chi_b = self.expe.get('chi_b', 1)
        self._tau_b = self.expe.get('tau_b', 100)
        self._kappa_b = self.expe.get('kappa_b', 0.9)
        self._update_gstep_theta()
        self._update_gstep_phi()

    def _init_gradienta(self):
        self._timestep_a = 0
        self._timestep_b = 0
        self._chi_a = self.expe.get('chi', 1)
        self._tau_a =  self.expe.get('tau', 42)
        self._kappa_a = self.expe.get('kappa', 0.75)
        self._chi_b = 42
        self._tau_b = 300
        self._kappa_b = 0.9
        self._update_gstep_theta()
        self._update_gstep_phi()

    def _init_gradientb(self):
        self._timestep_a = 0
        self._timestep_b = 0
        self._chi_a = 1
        self._tau_a =  42
        self._kappa_a = 0.75
        self._chi_b = self.expe.get('chi', 42)
        self._tau_b = self.expe.get('tau', 300)
        self._kappa_b = self.expe.get('kappa', 0.9)
        self._update_gstep_theta()
        self._update_gstep_phi()

    def _random_s_init(self):
        K = self._len['K']
        N = self._len['N']
        nfeat = self._len['nfeat']
        dims = self._len['dims']
        zeros = self._len['zeros']
        ones = self._len['ones']
        nnz = self._len['nnz']

        ##### Topic assignment
        dim = (N, N, 2)

        # Poisson way
        #alpha_0 = self.alpha_0
        #z = np.array( [poisson(alpha_0, size=dim) for dim in data_dims] )

        # Random way
        K = self._len['K']
        z = np.random.randint(0, K, (dim))

        if self._is_symmetric:
            z[:, :, 0] = np.triu(z[:, :, 0]) + np.triu(z[:, :, 0], 1).T
            z[:, :, 1] = np.triu(z[:, :, 1]) + np.triu(z[:, :, 1], 1).T

        ### Theta recovering
        theta_left = np.zeros((N, K), dtype=float)
        theta_right = np.zeros((N, K), dtype=float)

        self._symmetric_pt = self._is_symmetric +1
        #self._symmetric_pt = 1

        for i, j in self.data_iter(self.frontend.data_ma):
            k_i = z[i, j, 0]
            k_j = z[i, j, 1]
            theta_left[i, k_i] +=  self._symmetric_pt
            theta_right[j, k_j] += self._symmetric_pt

        ##### phi assignment
        phi = np.zeros((nfeat, K, K), dtype=float)

        for i, j in self.data_iter(self.frontend.data_ma):
            z_ij = z[i,j,0]
            z_ji = z[i,j,1]
            phi[self.frontend.data_ma[i, j], z_ij, z_ji] += 1
            if self._is_symmetric:
                phi[self.frontend.data_ma[j, i], z_ji, z_ij] += 1

        self.N_theta_left = theta_left
        self.N_theta_right = theta_right
        self.N_phi = phi

        # Temp Containers (for minibatch)
        self._N_phi = np.zeros((nfeat, K,K))
        self.hyper_phi_sum = self.hyper_phi.sum()
        self.hyper_theta_sum = self.hyper_theta.sum()

        return [self.N_phi, self.N_theta_right]

    def _random_ss_init(self):
        ''' Sufficient Statistics Initialization '''
        K = self._len['K']
        N = self._len['N']
        nfeat = self._len['nfeat']
        dims = self._len['dims']
        zeros = self._len['zeros']
        ones = self._len['ones']
        nnz = self._len['nnz']

        #N_theta_left = (dims[:, None] * np.random.dirichlet(np.ones(K), N))
        #N_theta_right = (dims[:, None] * np.random.dirichlet(np.ones(K), N))
        N_theta_left = (dims[:, None] * np.random.dirichlet([0.5]*K, N))
        N_theta_right = (dims[:, None] * np.random.dirichlet([0.5]*K, N))

        #N_phi = np.random.dirichlet([zeros, ones], K**2).T.reshape(2,K,K)
        #N_phi = nnz / (K**2) *  np.random.dirichlet([1, 1], K**2).T.reshape(2,K,K)

        N_phi = np.zeros((2,K,K))
        N_phi[0] = np.random.dirichlet([0.5]*K**2).reshape(K,K) * zeros
        N_phi[1] = np.random.dirichlet([0.5]*K**2).reshape(K,K) * ones

        self.N_phi = N_phi
        self.N_theta_left = N_theta_left
        self.N_theta_right = N_theta_right


        # Temp Containers (for minibatch)
        self._N_phi = np.zeros((nfeat, K,K))
        self.hyper_phi_sum = self.hyper_phi.sum()
        self.hyper_theta_sum = self.hyper_theta.sum()

        #self._qij = self.likelihood(*self._reduce_latent())
        self._symmetric_pt = self._is_symmetric +1

        # Return sufficient statistics
        return [N_phi, N_theta_left, N_theta_right]

    def _random_cvb_init(self):
        K = self._len['K']
        N = self._len['N']
        nfeat = self._len['nfeat']
        dims = self._len['dims']
        zeros = self._len['zeros']
        ones = self._len['ones']
        nnz = self._len['nnz']

        if self._is_symmetric and False:
            self.gamma = np.zeros((N,N,K,K))
            for i, j in self.data_iter():
                gmma = np.random.randint(1, N, (K,K))
                self.frontend.symmetrize(gmma)
                gmma = gmma / gmma.sum()
                self.gamma[i,j] = gmma

                #self.gamma[j,i] = gmma # data_iter 2
                gmma = np.random.randint(1, N, (K,K))
                self.frontend.symmetrize(gmma)
                gmma = gmma / gmma.sum()
                self.gamma[j,i] = gmma
        else:
            self.gamma = np.random.dirichlet(np.ones(K**2)*0.5, (N,N))
            self.gamma.resize(N,N,K,K)
            #self.gamma[self.frontend.data_ma == np.ma.masked] = 0 # ???

        #self._symmetric_pt = 1

        self.N_theta_left = self.gamma.sum(0).sum(1)
        self.N_theta_right = self.gamma.sum(1).sum(2)

        self.N_phi = np.zeros((nfeat, K, K))
        self.N_phi[0] = self.gamma[self.frontend.data_ma == 0].sum(0)
        self.N_phi[1] = self.gamma[self.frontend.data_ma == 1].sum(0)


    def _reset_containers(self):
        self._N_phi *= 0
        self.samples = []
        return

    def _is_container_empty(self):
        return len(self.samples) == 0

    def _update_gstep_theta(self):
        ''' Gradient converge for kappa _in (0.5,1] '''
        chi = self._chi_a
        tau = self._tau_a
        kappa = self._kappa_a

        self.gstep_theta = chi / ((tau + self._timestep_a)**kappa)


    def _update_gstep_phi(self):
        chi = self._chi_b
        tau = self._tau_b
        kappa = self._kappa_b

        self.gstep_phi =  chi / ((tau + self._timestep_b)**kappa)


    def _reduce_latent(self):
        theta = self.N_theta_right + self.N_theta_left + np.tile(self.hyper_theta, (self.N_theta_left.shape[0],1))
        self._theta = (theta.T / theta.sum(axis=1)).T

        phi = self.N_phi + np.tile(self.hyper_phi, (self.N_phi.shape[1], self.N_phi.shape[2], 1)).T
        #phi = (phi / np.linalg.norm(phi, axis=0))[1]
        self._phi = (phi / phi.sum(0))[1]

        self._K = self.N_phi.shape[1]

        return self._theta, self._phi

    def _reduce_one(self, i, j):
        xij = self._xij

        self.pik = self.N_theta_left[i] + self.hyper_theta
        self.pjk = self.N_theta_right[j] + self.hyper_theta
        pxk = self.N_phi[xij] + self.hyper_phi[xij]

        outer_kk = np.log(np.outer(self.pik, self.pjk)) + np.log(pxk) - np.log(self.N_phi.sum(0) + self.hyper_phi_sum)

        return lognormalize(outer_kk.ravel())


    def maximization(self, iter):
        ''' Variational Objective '''
        i,j = iter
        #self.pull_current(i, j)
        variational = self._reduce_one(i,j).reshape(self._len['K'], self._len['K'])
        #self.push_current(i, j, variational)
        self.samples.append(variational)

    def expectation(self, iter, burnin=False):
        ''' Follow the White Rabbit '''
        i,j = iter
        xij = self._xij
        qij = self.samples[-1]

        self._update_local_gradient(i, j, qij)
        self._timestep_a += 1

        if burnin:
            self.samples = []
            return
        else:
            self._update_global_gradient(i, j, qij, xij)
            #self._purge_minibatch()
            #self._timestep_b += 1
            pass


    def _update_local_gradient(self, i, j, qij):
        _len = self._len

        # Sum ?
        q_left = qij.sum(0)
        q_right = qij.sum(1)


        self.N_theta_left[i] = (1 - self.gstep_theta)*self.N_theta_left[i] + (self.gstep_theta * _len['dims'][i] * q_left)
        self.N_theta_right[j] = (1 - self.gstep_theta)*self.N_theta_right[j] + (self.gstep_theta * _len['dims'][j] * q_right)

        if self._is_symmetric:
            self.N_theta_left[j] = (1 - self.gstep_theta)*self.N_theta_left[j] + (self.gstep_theta * _len['dims'][j] * q_left)
            self.N_theta_right[i] = (1 - self.gstep_theta)*self.N_theta_right[i] + (self.gstep_theta * _len['dims'][i] * q_right)


        # or Mean ?
        #self.N_theta_left[i]  = (1 - self.gstep_theta)*self.N_theta_left[i]  + (self.gstep_theta * _len['dims'][i] * qij.mean(1))
        #self.N_theta_right[j] = (1 - self.gstep_theta)*self.N_theta_right[j] + (self.gstep_theta * _len['dims'][j] * qij.mean(0))

        # or Right  ?
        #pik = self.pik / (2*len(_len['dims']) + self.hyper_theta_sum)
        #pjk = self.pjk / (2*len(_len['dims']) + self.hyper_theta_sum)
        #self.N_theta_left[i]  = (1 - self.gstep_theta)*self.N_theta_left[i]  + (self.gstep_theta * _len['dims'][i] * pik)
        #self.N_theta_right[j] = (1 - self.gstep_theta)*self.N_theta_right[j] + (self.gstep_theta * _len['dims'][j] * pjk)

        self._update_gstep_theta()

    def _update_global_gradient(self, i, j, qij, xij):
        self._N_phi[xij] += qij * self._symmetric_pt


    def _purge_minibatch(self):
        ''' Update the global gradient then purge containers '''
        if not self._is_container_empty():
            self.N_phi = (1 - self.gstep_phi)*self.N_phi + self.gstep_phi * (self._len['nnz'] / len(self.samples)) * self._N_phi

        self._update_gstep_phi()

        self._reset_containers()

        self._timestep_b += 1


    def compute_entropy(self):
        pij = self.likelihood(*self._reduce_latent())

        # Log-likelihood
        pij = self.frontend.data_A * pij + self.frontend.data_B
        #ll = np.log(pij).sum()
        ll = np.log(pij).sum()

        # Entropy
        entropy = ll

        #entropy = - ll / self._len['nnz']

        # Perplexity is 2**H(X).

        return entropy

    def compute_entropy_t(self):
        pij = self.likelihood(*self._reduce_latent())

        # Log-likelihood
        pij = self.frontend.data_A_t * pij + self.frontend.data_B_t
        ll = np.log(pij).sum()

        # Entropy
        entropy_t = ll

        #entropy_t = - ll / self._len['nnz_t']
        # Perplexity is 2**H(X).

        return entropy_t

    def compute_elbo(self):

        pij = self.likelihood(*self._reduce_latent())

        pij = ma.array(pij, mask=self.get_mask())

        # Log-likelihood
        ll = self.frontend.data_A * pij + self.frontend.data_B


        eq = (pij * np.log(ll)).sum()
        hq = (pij * np.log(pij)).sum()

        elbo = eq - hq

        return elbo


    def update_hyper(self, hyper):
        pass

    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        #self.update_hyper(hyperparams)
        #alpha, gmma, delta = self.get_hyper()

        # predictive
        try: theta, phi = self.get_params()
        except: return self.generate(N, K, hyperparams, 'generative', symmetric)
        K = theta.shape[1]

        pij = self.likelihood(theta, phi)
        pij = np.clip(pij, 0, 1)
        Y = sp.stats.bernoulli.rvs(pij)

        return Y



if __name__ == "__main__":

    import pymake
    from pymake import frontendNetwork

    data = np.arange(16).reshape(4,4)
    data_ma = np.ma.array(data, mask = data*0)

    data_ma.mask[0,1] = True
    data_ma.mask[1,1] = True
    data_ma.mask[1,2] = True
    data_ma.mask[1,3] = True
    data_ma.mask[3,3] = True

    fr = frontendNetwork.from_array(data_ma)

    model = immsb_scvb({}, fr)

    _abc = model.data_iter(data_ma)

    # data to iter
    print(data_ma[zip(*_abc)])




import os, re
from string import punctuation
from pymake.util.utils import basestring


import numpy as np
import scipy as sp


class Vocabulary(object):

    recover_list = {"wa":"was", "ha":"has"}

    regex = {
        'word' : re.compile(r'[a-zA-Z0-9_\-]+'),
        'not_word' : re.compile(r'[^a-zA-Z0-9_\-]+'),
        'ponctuation' : re.compile(r'[\.\,\;\:\!\?\(\)\{\}\[\]\'\"\`]+'),
    }

    def __init__(self, exclude_stopwords=False, lemmatize=True):

        try:
            import nltk
            _NLTK_DISABLED = False
        except:
            _NLTK_DISABLED = True

        self.vocas = []        # id to word
        self.token2id = dict() # word to id
        self.docfreq = []      # id to document frequency
        self.exclude_stopwords = exclude_stopwords

        stopwords_list = []
        if exclude_stopwords:
            # Too much strict
            #with open (os.path.join(os.path.dirname(__file__), 'stopwords.txt'), "r") as _f:
            #    stopwords_list = _f.read().replace('\n', '').split()
            if not _NLTK_DISABLED:
                stopwords_list += nltk.corpus.stopwords.words('english')
            stopwords_list = set(stopwords_list)
        self.stopwords_list = stopwords_list

        if lemmatize:
            if not _NLTK_DISABLED:
                self.wlemm = nltk.WordNetLemmatizer()
            else:
                print ('Warning: no lemmatizer !')


    def is_stopword(self, w):
        return w in self.stopwords_list

    def lemmatize(self, w0):
        if not hasattr(self, 'wlemm'):
            #self.log.debug()
            print('No lematization')
            return w0
        w = self.wlemm.lemmatize(w0.lower())
        if w in self.recover_list: return self.recover_list[w]
        return w

    def token2id(self):
        return self.token2id

    def id2token(self):
        if hasattr(self, '_id2token') and len(self.token2id) == len(self._id2token):
            return self._id2token
        else:
            self._id2token = dict((v,k) for k, v in self.token2id.iteritems())
            return self._id2token

    def term_to_id(self, term0):
        term = self.lemmatize(term0)
        if not self.regex['word'].match(term): return None
        if self.exclude_stopwords and self.is_stopword(term): return None
        try:
            term_id = self.token2id[term]
        except:
            term_id = len(self.vocas)
            self.token2id[term] = term_id
            self.vocas.append(term)
            self.docfreq.append(0)
        return term_id

    def remove_stopwords(self, doc):
        if isinstance(doc, basestring):
            doc = doc.split()
        elif doc is None:
            return None
        else:
            doc = doc
        return ' '.join([w for w in doc if not self.is_stopword(w)]).strip()

    # Bag of words !
    def doc2bow(self, doc):
        l = dict()
        words = dict()
        doc = self.regex['ponctuation'].sub(' ', doc).split() if isinstance(doc, basestring) else doc
        for term in doc:
            id = self.term_to_id(term)
            if id != None:
                l[id] = l.get(id, 0) + 1
                if not id in words:
                    words[id] = 1
                    self.docfreq[id] += 1 # It counts in how many documents a word appears. If it appears in only a few, remove it from the vocabulary using cut_low_freq()
        if "close" in dir(doc): doc.close()
        return sorted(l.items())

    def cut_low_freq(self, corpus, threshold=1):
        new_vocas = []
        new_docfreq = []
        self.token2id = dict()
        conv_map = dict()
        for id, term in enumerate(self.vocas):
            freq = self.docfreq[id]
            if freq > threshold:
                new_id = len(new_vocas)
                self.token2id[term] = new_id
                new_vocas.append(term)
                new_docfreq.append(freq)
                conv_map[id] = new_id
        self.vocas = new_vocas
        self.docfreq = new_docfreq

        def conv(doc):
            new_doc = []
            for id in doc:
                if id in conv_map: new_doc.append(conv_map[id])
            return new_doc
        return [conv(doc) for doc in corpus]

    def __getitem__(self, v):
        return self.vocas[v]

    def size(self):
        return len(self.vocas)

    def is_stopword_id(self, id):
        return self.vocas[id] in self.stopwords_list

    # if string in input
    #   * return a corpus as list of list of string.
    # Else
    #   * return a numpy array of cooc-matrix
    @classmethod
    def parse_corpus(cls, fname, bdir=""):
        dico = None
        if type(fname) is str:
            if not os.path.exists(fname): raise EnvironmentError('%s ?' % fname)
            if os.path.isdir(fname):
                import fnmatch
                bdir = fname
                corpus_files = []
                dico_files = []
                for root, dirnames, filenames in os.walk(bdir):
                    for filename in fnmatch.filter(filenames, '*.txt'):
                        if filename.startswith(('dico.','vocab.')):
                            dico_files.append(os.path.join(root, filename))
                        else:
                            corpus_files.append(os.path.join(root, filename))

                if len(corpus_files) == 1:
                    # Parse sparse matrix:  DOC_ID WORD_ID COUNT
                    with open(corpus_files[0],'r') as f:
                        # Read the header information
                        n_instances = int(f.readline())
                        n_features = int(f.readline())
                        n_nnz = int(f.readline())

                        # Create cooc-matrix
                        #data = sp.sparse.csr_matrix((n_instances, n_features))
                        data = sp.sparse.lil_matrix((n_instances, n_features), dtype=int)
                        for line in f:
                            doc_id, word_id, count = list(map(int, line.split()))
                            data[doc_id-1, word_id-1] = count
                        data = data.tocsr()
                    # If dictionnary, return it
                    with open(dico_files[0],'r') as f:
                        dico = {}
                        for i, line in enumerate(f):
                            word = line.split()
                            assert(len(word) == 1)
                            #dict((v,k) for k, v in self.token2id.iteritems())
                            dico[i] = word[0]
                else:
                    # Parse documents as each of them is a .txt file
                    data = [cls.parse_document_f(f) for f in corpus_files]

            #elif os.path.isfile(fname): ?
        elif type(fname) is list:
            # List of string as document
            docs = fname
            data = [cls.parse_document_l(d) for d in docs]
        else:
            raise NotImplementedError('file input: %s' % fname)

        return data, dico

    @classmethod
    def parse_document_f(cls, d):
        return [word.lower() for line in open(d).readlines() for word in cls.regex['not_word'].sub(' ', line).split()]

    @classmethod
    def parse_document_l(cls, d):
        return [word.lower() for word in cls.regex['not_word'].sub(' ', d).split()]






from functools import partial
from pymake.core.gram import exp_append, exp_append_uniq

_gram = [
    #  Expe Settings -- Context-Free
    #
    #  * Are repeatable



    '--epoch', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help='number for averaginf generative process'),

    '-c','--corpus', dict(
        nargs='*', dest='corpus', action=exp_append,
        help='ID of the frontend data.'),

    '-r','--random', dict(
        nargs='*', dest='corpus', action=exp_append,
        help='Random generation of synthetic frontend  data [uniforma|alternate|cliqueN|BA].'),

    '-m','--model', dict(
        nargs='*',dest='model', action=exp_append,
        help='ID of the model.'),

    '-n','--N', dict(
        nargs='*', action=exp_append, # str because keywords "all"
        help='Size of frontend data [int | all].'),

    '-k','--K', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help='Latent dimensions'),

    '-i','--iterations', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help='Max number of iterations for the optimization.'),

    '--hyper',  dict(
        dest='hyper', nargs='*', action=exp_append,
        help='type of hyperparameters optimization [auto|fix|symmetric|asymmetric]'),

    '--hyper-prior','--hyper_prior', dict(
        dest='hyper_prior', action=partial(exp_append_uniq, _t=float), nargs='*',
        help='Set paramters of the hyper-optimization [auto|fix|symmetric|asymmetric]'),

    '--testset-ratio', dict(
        dest='testset_ratio', nargs='*', action=partial(exp_append, _t=float),
        help='testset/learnset percentage for testing.'),

    '--mask', dict(
        nargs='*', action=exp_append,
        help='mask type (balanced|unbalanced'),

    '--homo', dict(
        nargs='*', action=exp_append,
        help='Centrality type (NotImplemented)'),

    '--alpha', dict(
        nargs='*', action=exp_append,
        help='First hyperparameter.'),
    '--gmma', dict(
        nargs='*', action=exp_append,
        help='Second hyperparameter.'),
    '--delta', dict(
        nargs='*', action=exp_append,
        help='Third hyperparameter.'),
    '--chunk', dict(
        nargs='*', action=partial(exp_append, _t=str),
        help='Chunk size for online learning.'),
    '--burnin', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help='Number of samples used for burnin period.'),

    # step for gradient
    '--chi', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help=''),
    '--tau', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help=''),
    '--kappa', dict(
        nargs='*', action=partial(exp_append, _t=int),
        help=''),


    # System
     '--snapshot',dict(dest='snapshot_freq', type=int),

     '--driver', dict(
         help='Choose the driver to use to load data frontend.'),

    '-g', '--generative',dict(dest='_mode',
                              action='store_const', const='generative'),
    '-p', '--predictive', dict(dest='_mode',
                               action='store_const', const='predictive'),


]



import os
import logging

from pymake import GramExp



class DataBase(object):
    """ Root Class for Frontend Manipulation over Corpuses and Models.

        Given Data Y, and Model M = {\theta, \Phi}
        E[Y] = \theta \phi^T

        Fonctionality are of the frontend decline as:
        1. Frontend for model/algorithm I/O,
        2. Frontend for Corpus Information, and Result Gathering for
            Machine Learning Models.
        3. Data Analisis and Prediction..

        load_corpus -> load_text_corpus -> text_loader
        (frontent)  ->   (choice)       -> (adapt preprocessing)

    """

    log = logging.getLogger('root')

    def __init__(self, expe):
        self.expe = expe

        self._force_load_data = expe.get('_force_load_data', True)
        self._force_save_data = expe.get('_force_save_data', True)

        self.corpus_name = expe.get('corpus')

    #
    # I/O Methods
    #

    @classmethod
    def from_expe(cls):
        raise NotImplementedError

    @classmethod
    def _extract_data(cls):
        ''' Raw data parsing/extraction. '''
        raise NotImplementedError

    @staticmethod
    def get_input_path(expe):
        if '_input_path' in expe:
            return expe['_input_path']
        else:
            input_path =  GramExp.make_input_path(expe)
            expe['_input_path'] = input_path
            return input_path


    @classmethod
    def _resolve_filename(cls, expe):
        input_path = expe._input_path

        if not os.path.exists(input_path):
            cls.log.error("Corpus `%s' Not found." % (input_path))
            print('please run "fetch_networks"')
            self.data = None
            return

        if expe.corpus.endswith('.pk'):
            basename = expe.corpus
        else:
            basename = expe.corpus + '.pk'

        fn = os.path.join(input_path, basename)
        return fn

    @classmethod
    def _load_data(cls, *args, **kwargs):
        ''' Load preprocessed data. '''
        from pymake.io import load
        return load(*args, **kwargs)

    @classmethod
    def _save_data(cls, *args, **kwargs):
        ''' Save preprocessed data. '''
        from pymake.io import save
        return save(*args, **kwargs)

    def configure(self):
        ''' Configure the frontend Data.
        Try the following steps:
            1. Sample the corpus (expe.N),
            2. Build a testset/validation set (expe.testset_ratio & mask),
         '''
        if self.data is None:
            return

        if self.expe.get('exponentiate'):
            MAX = 300
            self.data.ep['weights'].a = 2**self.data.ep['weights'].a
            self.data.ep['weights'].a[self.data.ep['weights'].a > MAX] = int(MAX)
            self.data.ep['weights'].a[self.data.ep['weights'].a < 0] = int(MAX)

        N = self.expe.get('N')
        if N is not None and N != 'all':
            N = int(N)
            self.log.debug('sampling dataset to N=%d ...' % N)
            self.sample(N)

        testset_ratio = self.expe.get('testset_ratio')
        if testset_ratio is not None:
            self.log.debug('Building testset ...')
            self.make_testset()


        if self.expe.get('noise'):
            self.make_noise()

        return





from numpy import ma
import numpy as np
import logging

try:
    import pandas as pd
except Exception as e:
    print('Error while importing pandas: %s' % e)



class DatasetDriver(object):

    ''' Parse dataset file using pandas'''

    _comment = '%'
    log = logging.getLogger('root')

    # No pandas here....
    @classmethod
    def parse_tnet(cls, fn, sep=' '):
        ''' Grammar retro-ingennired from fb/emaileu.txt. tnet format is official ? '''
        cls.log.debug('opening file: %s' % fn)
        with open(fn) as f:
            content = f.read()
        lines = list(filter(None, content.split('\n')))
        line1_length = lines[0].strip().split(sep)
        edges = {}
        if len(line1_length) == 2:
            # format 'i j' if edges.
            data_file_format = 'txt'
            for line in lines:
                dyad = line.strip().split(sep)
                dyad = '.'.join(dyad)
                edges[dyad] = edges.get(dyad, 0) + 1
            #edges = [l.strip().split(sep) for l in lines]
        elif len(line1_length) == 5:
            # format '"date" i j weight'.
            data_file_format = 'tnet'
            for line in lines:
                _line = line.strip().split(sep)
                dyad = _line[-3:-1]
                dyad = '.'.join(dyad)
                w = int(_line[-1])
                edges[dyad] = edges.get(dyad, 0) + w
            #edges = [l.strip().split(sep)[-3:-1] for l in lines]

        edges = np.array([ (e.split('.')[0], e.split('.')[1], w+1) for e, w in edges.items()], dtype=int) -1

        edges[:, 0:2] -= edges[:, 0:2].min()
        N = edges[:, 0:2].max()+1

        g = np.zeros((N,N))
        g[tuple(edges[:, :2].T)] = edges[:, 2]
        data = dict(data=g)
        return data

    # No pandas here....
    @classmethod
    def parse_csv(cls, fn, sep=';'):
        ''' Grammar retro-ingennired from manufacturing.csv '''
        cls.log.debug('opening file: %s' % fn)
        with open(fn, 'r') as f:
            content = f.read()
        lines = list(filter(None, content.split('\n')))[1:]
        edges = {}
        for line in lines:
            dyad = line.strip().split(sep)[0:2]
            dyad = '.'.join(dyad)
            edges[dyad] = edges.get(dyad, 0) + 1
        #edges = [l.strip().split(sep)[0:2] for l in lines]
        #edges = np.array([ (e[0], e[1]) for e in edges], dtype=int) -1
        edges = np.array([ (e.split('.')[0], e.split('.')[1], w+1) for e, w in edges.items()], dtype=int) -1

        edges[:, 0:2] -= edges[:, 0:2].min()
        N = edges[:, 0:2].max()+1

        g = np.zeros((N,N))
        g[tuple(edges[:, :2].T)] = edges[:, 2]
        data = dict(data=g)
        return data


    @classmethod
    def parse_dancer(cls, fn, sep=';'):
        """ Parse Network data depending on type/extension """
        cls.log.debug('opening file: %s' % fn)

        data = pd.read_csv(fn, sep=sep, names=['n', 'feat', 'cluster' ], comment=cls._comment)
        parameters = data.dropna()
        clusters = parameters['cluster'].values.astype(int)
        features = np.array([list(map(float, f.split('|'))) for f in parameters['feat'].values])

        data = data.ix[data['cluster'].isna()]
        data['cluster'] = 1 # <= the weight
        data = data.loc[pd.to_numeric(data['n'], errors='coerce').dropna().index].as_matrix().astype(int)


        data[:, 0:2] -= data[:, 0:2].min()
        N = data[:, 0:2].max()+1
        y = np.zeros((N,N))
        e_l = data[:,2] > 0
        e_ix = data[:, 0:2][e_l]
        ix = list(zip(*e_ix))
        y[ix] = data[:,2][e_l]

        data = dict(data=y, clusters=clusters, features=features)
        return data

    @classmethod
    def parse_dat(cls, fn, sep="\s+"):
        """ Parse Network data depending on type/extension """
        cls.log.debug('opening file: %s' % fn)

        def _row_len(fn):
            ''' Seek for the length of the csv row, then break quicly '''
            inside = {'vertices':False, 'edges':False }
            data  = []
            for _line in open(fn):
                line = _line.strip()
                if line.startswith(('ROW LABELS:', '*vertices')) or inside['vertices']:
                    if not inside['vertices']:
                        inside['vertices'] = True
                        continue
                    if line.startswith('#') or not line.strip():
                        inside['vertices'] = False # break
                    elif line.startswith(('DATA','*edges' )):
                        inside['vertices'] = False # break
                        inside['edges'] = True
                    else:
                        continue
                elif line.startswith(('DATA','*edges' )) or inside['edges']:
                    if not inside['edges']:
                        inside['edges'] = True # break
                        continue
                    if line.startswith('#') or not line.strip() or len(line.split()) < 2 :
                        inside['edges'] = False
                    else:
                        # Parsing assignation
                        data.append( line.split() )
                        break
            return len(data[0])

        # Sender, Reiceiver, Edges
        row_len = _row_len(fn)
        if  row_len == 3:
            cols = ['s', 'r', 'weight']
        elif row_len == 2:
            cols = ['s', 'r']
        else:
            raise ValueError('I/O error for dataset file: %s' % fn)

        data = pd.read_csv(fn, sep=sep, names=cols, comment=cls._comment)
        if len(cols) == 2:
            data['weight'] = np.ones(data.shape[0])
            cols = ['s', 'r', 'weight']

        cond = pd.to_numeric(data['s'], errors='coerce').dropna().index & pd.to_numeric(data['r'], errors='coerce').dropna().index
        data = data.loc[cond].as_matrix().astype(int)


        data[:, 0:2] -= data[:, 0:2].min()
        N = data[:, 0:2].max()+1
        y = np.zeros((N,N))
        e_l = data[:,2] > 0
        e_ix = data[:, 0:2][e_l]
        ix = list(zip(*e_ix))
        y[ix] = data[:,2][e_l]
        data = dict(data=y)
        return data



class OnlineDatasetDriver(object):

    ''' Parse dataset file using pandas'''

    _comment = '%'
    log = logging.getLogger('root')

    @classmethod
    def parse_tnet(cls, fn, sep=' '):
        ''' Grammar retro-ingennired from fb/emaileu.txt. tnet format is official ? '''

        cls.log.debug('opening file: %s' % fn)

        for line in open(fn):
            line = line.strip()
            if not line:
                continue

            line1_length = line.split(sep)

            if len(line1_length) == 2:
                # format 'i j' if edges.
                data_file_format = 'txt'
                v1, v2 = line.strip().split(sep)
                w = 1
                yield int(v1), int(v2), w, None
            elif len(line1_length) == 5:
                # format '"date" i j weight'.
                data_file_format = 'tnet'
                _line = line.strip().split(sep)
                v1, v2 = _line[-3:-1]
                w = int(_line[-1])
                if w == 0:
                    continue
                else:
                    yield int(v1), int(v2), w, None


    @classmethod
    def parse_csv(cls, fn, sep=';'):
        ''' Grammar retro-ingennired from manufacturing.csv '''

        cls.log.debug('opening file: %s' % fn)

        cpt = 0
        for line in open(fn):
            if cpt == 0:
                # Ignore first status line
                cpt += 1
                continue
            v1, v2 = line.strip().split(sep)[0:2]
            w = 1
            yield int(v1), int(v2), w, None


    @classmethod
    def parse_dancer(cls, fn, sep=';'):

        cls.log.debug('opening file: %s' % fn)

        inside = {'vertices':False, 'edges':False }
        for line in open(fn):
            line = line.strip()
            if line.startswith('# Vertices') or inside['vertices']:
                if not inside['vertices']:
                    inside['vertices'] = True
                    continue
                if line.startswith('#') or not line.strip() :
                    inside['vertices'] = False # break
                else:
                    # Parsing assignation
                    elements = line.strip().split(sep)
                    index = int(elements[0])
                    clust = int(elements[-1])
                    feats = list(map(float, elements[-2].split('|')))
                    obj = {'cluster': clust, 'features': feats, 'index':index}
                    yield obj
            elif line.startswith('# Edges') or inside['edges']:
                if not inside['edges']:
                    inside['edges'] = True
                    continue
                if line.startswith('#') or not line.strip() :
                    inside['edges'] = False # break
                else:
                    # Parsing assignation
                    v1, v2 = line.split(sep)
                    w = 1
                    yield int(v1), int(v2), w, None


    @classmethod
    def parse_dat(cls, fn, sep=" "):
        """ Parse Network data depending on type/extension """

        cls.log.debug('opening file: %s' % fn)

        inside = {'vertices':False, 'edges':False }
        for line in open(fn):
            line = line.strip()
            if line.startswith(('ROW LABELS:', '*vertices')) or inside['vertices']:
                if not inside['vertices']:
                    inside['vertices'] = True
                    continue
                if line.startswith('#') or not line.strip():
                    inside['vertices'] = False # break
                elif line.startswith(('DATA','*edges' )):
                    inside['vertices'] = False # break
                    inside['edges'] = True
                else:
                    continue
            elif line.startswith(('DATA','*edges' )) or inside['edges']:
                if not inside['edges']:
                    inside['edges'] = True # break
                    continue
                if line.startswith('#') or not line.strip() or len(line.split()) < 2 :
                    inside['edges'] = False
                else:
                    # Parsing assignation
                    splitline = line.split(sep)
                    row_size = len(splitline)
                    if row_size == 2:
                        # like .txt
                        v1, v2 = splitline
                        w = 1
                        yield int(v1), int(v2), w, None
                    elif row_size == 3:
                        v1, v2 = splitline[0:2]
                        w = int(splitline[2])
                        if w == 0:
                            continue
                        else:
                            yield int(v1), int(v2), w, None
                    else:
                        raise NotImplementedError








class RawDatasetDriver(object):

    ''' Parse dataset file using python loop (deprecated) '''

    _comment = '%'
    log = logging.getLogger('root')

    @classmethod
    def parse_tnet(cls, fn, sep=' '):
        ''' Grammar retro-ingennired from fb/emaileu.txt '''
        cls.log.debug('opening file: %s' % fn)
        with open(fn) as f:
            content = f.read()
        lines = list(filter(None, content.split('\n')))
        line1_length = lines[0].strip().split(sep)
        edges = {}
        if len(line1_length) == 2:
            # format 'i j' if edges.
            data_file_format = 'txt'
            for line in lines:
                dyad = line.strip().split(sep)
                dyad = '.'.join(dyad)
                edges[dyad] = edges.get(dyad, 0) + 1
            #edges = [l.strip().split(sep) for l in lines]
        elif len(line1_length) == 5:
            # format '"date" i j weight'.
            data_file_format = 'tnet'
            for line in lines:
                _line = line.strip().split(sep)
                dyad = _line[-3:-1]
                dyad = '.'.join(dyad)
                w = int(_line[-1])
                edges[dyad] = edges.get(dyad, 0) + w
            #edges = [l.strip().split(sep)[-3:-1] for l in lines]

        edges = np.array([ (e.split('.')[0], e.split('.')[1], w+1) for e, w in edges.items()], dtype=int) -1
        N = edges.max() +1
        #N = max(list(itertools.chain(*edges))) + 1

        g = np.zeros((N,N))
        g[tuple(edges[:, :2].T)] = edges[:, 2]
        data = dict(data=g)
        return data

    @classmethod
    def parse_csv(cls, fn, sep=';'):
        ''' Grammar retro-ingennired from manufacturing.csv '''
        cls.log.debug('opening file: %s' % fn)
        with open(fn) as f:
            content = f.read()
        lines = list(filter(None, content.split('\n')))[1:]
        edges = {}
        for line in lines:
            dyad = line.strip().split(sep)[0:2]
            dyad = '.'.join(dyad)
            edges[dyad] = edges.get(dyad, 0) + 1
        #edges = [l.strip().split(sep)[0:2] for l in lines]
        #edges = np.array([ (e[0], e[1]) for e in edges], dtype=int) -1
        edges = np.array([ (e.split('.')[0], e.split('.')[1], w+1) for e, w in edges.items()], dtype=int) -1
        N = edges.max() +1
        #N = max(list(itertools.chain(*edges))) + 1

        g = np.zeros((N,N))
        g[tuple(edges[:, :2].T)] = edges[:, 2]
        data = dict(data=g)
        return data

    @classmethod
    def parse_dancer(cls, fn, sep=';'):
        """ Parse Network data depending on type/extension """
        cls.log.debug('opening file: %s' % fn)
        data = []
        inside = {'vertices':False, 'edges':False }
        clusters = []
        features = []
        for line in open(fn):
            if line.startswith('# Vertices') or inside['vertices']:
                if not inside['vertices']:
                    inside['vertices'] = True
                    N = 0
                    continue
                if line.startswith('#') or not line.strip() :
                    inside['vertices'] = False # break
                else:
                    # Parsing assignation
                    elements = line.strip().split(sep)
                    clust = int(elements[-1])
                    feats = list(map(float, elements[-2].split('|')))
                    clusters.append(clust)
                    features.append(feats)
                    N += 1
            elif line.startswith('# Edges') or inside['edges']:
                if not inside['edges']:
                    inside['edges'] = True
                    continue
                if line.startswith('#') or not line.strip() :
                    inside['edges'] = False # break
                else:
                    # Parsing assignation
                    data.append( line.strip() )

        edges = np.array([tuple(row.split(sep)) for row in data]).astype(int)
        g = np.zeros((N,N))
        g[[e[0] for e in edges], [e[1] for e in edges]] = 1
        g[[e[1] for e in edges], [e[0] for e in edges]] = 1
        # ?! .T

        try:
            parameters = parse_file_conf(os.path.join(os.path.dirname(fn), 'parameters'))
            parameters['devs'] = list(map(float, parameters['devs'].split(sep)))
        except IOError:
            parameters = {}
        finally:
            # @Obsolete !
            parameters_ = parameters

        clusters = clusters
        features = np.array(features)
        data = dict(data=g, clusters=clusters, features=features)
        return data

    @classmethod
    def parse_dat(cls, fn, sep=' '):
        """ Parse Network data depending on type/extension """
        cls.log.debug('opening file: %s' % fn)
        data = []
        inside = {'vertices':False, 'edges':False }
        for _line in open(fn):
            line = _line.strip()
            if line.startswith(('ROW LABELS:', '*vertices')) or inside['vertices']:
                if not inside['vertices']:
                    inside['vertices'] = True
                    continue
                if line.startswith('#') or not line.strip():
                    inside['vertices'] = False # break
                elif line.startswith(('DATA','*edges' )):
                    inside['vertices'] = False # break
                    inside['edges'] = True
                else:
                    # todo if needed
                    continue
            elif line.startswith(('DATA','*edges' )) or inside['edges']:
                if not inside['edges']:
                    inside['edges'] = True # break
                    continue
                if line.startswith('#') or not line.strip() or len(line.split(sep)) < 2 :
                    inside['edges'] = False
                else:
                    # Parsing assignation
                    data.append( line.strip() )

        row_size = len(data[0].split(sep))
        edges = np.array([tuple(row.split(sep)) for row in data]).astype(int)-1
        edges = {}
        if row_size == 2:
            # like .txt
            for line in data:
                dyad = line.strip().split(sep)
                dyad = '.'.join(dyad)
                edges[dyad] = edges.get(dyad, 0) + 1
        elif row_size == 3:
            for line in data:
                _line = line.strip().split(sep)
                dyad = _line[0:2]
                dyad = '.'.join(dyad)
                w = int(_line[-1]) # can be zeros
                edges[dyad] = edges.get(dyad, 0) + int(w)
        else:
            raise NotImplementedError

        edges = np.array([ (e.split('.')[0], e.split('.')[1], w+1) for e, w in edges.items()], dtype=int) -1
        N = edges.max() +1
        g = np.zeros((N,N))
        g[tuple(edges[:, :2].T)] = edges[:, 2]
        data = dict(data=g)
        return data




import os
if not os.environ.get('DISPLAY'):
    # Plot in nil/void
    import matplotlib; matplotlib.use('Agg')
    print("==> Warning : Unable to load DISPLAY, try : `export DISPLAY=:0.0'")
else:
    # Plot config
    __plot_font_size = 14
    import matplotlib.pyplot as plt
    plt.rc('font', size=__plot_font_size)  # controls default text sizes


# Expose the settings getter function
#from pymake.util.utils import get_pymake_settings
from pymake.core import get_pymake_settings

from pymake.core.types import Spec, Corpus, Model, Script, ExpSpace, ExpVector, ExpTensor, ExpDesign, ExpGroup
from pymake.core.format import ExpeFormat

# Without whoosh fashion...
#from pymake.frontend.io import SpecLoader
#__spec = SpecLoader.get_atoms()

from pymake.core.gramexp import GramExp





# -*- coding: utf-8 -*-

from time import time
import itertools
import logging
lgg = logging.getLogger('root')

import numpy as np
from numpy import ma
import scipy as sp

from scipy.special import digamma
from numpy.random import dirichlet, gamma, poisson, binomial, beta

from ml.model.modelbase import GibbsSampler
from .hdp import MSampler, BetaSampler

from pymake.util.math import lognormalize, categorical, sorted_perm, adj_to_degree, gem

# Implementation Mixed Membership Sochastic Blockmodel Stochastic

#import warnings
#warnings.simplefilter('error', RuntimeWarning)

""" @Todo
* sparse masked array ?!@$*

* Add constant to count matrix by default will win some precious second

ref:
HDP: Teh et al. (2005) Gibbs sampler for the Hierarchical Dirichlet Process: Direct assignement.
"""


# Broadcast class based on numpy matrix
# Assume delta a scalar.
class Likelihood(object):

    def __init__(self, delta, fr, nodes_list=None, assortativity=False):
        """ Notes
            -----
            * Diagonal is ignored here for prediction
        """
        if fr is None:
            return

        self.data_ma = fr.data_ma
        self.data_dims = self.get_data_dims()
        self.nnz = self.get_nnz()
        # Vocabulary size
        self.nfeat = self.get_nfeat() or 0

        if nodes_list is None:
            self.nodes_list = [np.arange(self.data_ma.shape[0]), np.arange(self.data_ma.shape[1])]
        else:
            self.nodes_list = nodes_list
            raise ValueError('re order the networks ! to avoid using _nmap')

        # assert for coocurence matric
        #assert(self.data_mat.shape[1] == self.nfeat)
        #self.data_mat = sppy.csarray(self.data_mat)

        # Cst for CGS of DM and scala delta as prior.
        self.assortativity = assortativity
        self.delta = np.asarray(delta) if isinstance(delta, (np.ndarray, list, tuple)) else np.asarray([delta] * self.nfeat)
        self.w_delta = self.delta.sum()
        if assortativity == 1:
            raise NotImplementedError('assort 2')
        elif assortativity == 2:
            self.epsilon = 0.01
        else:
            pass

        fr._set_rawdata_for_likelihood_computation()
        hook = ['data_A', 'data_B', 'data_A_t', 'data_B_t']
        for s in hook:
            setattr(self, s, getattr(fr, s))


    def compute(self, j, i, k_ji, k_ij):
        # _reduce_one !
        return self.loglikelihood(j, i, k_ji, k_ij)

    def get_nfeat(self):
        nfeat = self.data_ma.max() + 1
        if nfeat == 1:
            lgg.warn( 'Warning, only zeros in adjacency matrix...')
            nfeat = 2
        return nfeat

    # Contains the index of nodes with who it interact.
    # @debug no more true for bipartite networks
    def get_data_dims(self):
        #data_dims = np.vectorize(len)(self.data)
        #data_dims = [r.count() for r in self.data_ma]
        data_dims = []
        for i in range(self.data_ma.shape[0]):
            data_dims.append(self.data_ma[i,:].count() + self.data_ma[:,i].count())
        return data_dims

    def get_nnz(self):
        return len(self.data_ma.compressed())

    # Need it in the case of sampling sysmetric networks. (only case where ineed to map ?)
    # return the true node index corresponding to arbitrary index i of matrix count/data position
    # @pos: 0 indicate line picking, 1 indicate rows picking
    def _nmap(self, i, pos):
        return self.nodes_list[pos][i]

    # @debug: symmetric matrix ?
    def make_word_topic_counts(self, z, K):
        word_topic_counts = np.zeros((self.nfeat, K, K), dtype=int)

        for j, i in self.data_iter(randomize=True):
            z_ji = z[j,i,0]
            z_ij = z[j,i,1]
            word_topic_counts[self.data_ma[j, i], z_ji, z_ij] += 1
            if self._symmetric:
                word_topic_counts[self.data_ma[j, i], z_ij, z_ji] += 1

        self.word_topic_counts = word_topic_counts

    # Interface to properly iterate over data
    def data_iter(self, randomize=True):
        if not hasattr(self, '_order'):
            order = np.arange(self.data_ma.size).reshape(self.data_ma.shape)
            masked = order[self.data_ma.mask]

            if self._symmetric:
                tril = np.tril_indices_from(self.data_ma, -1)
                tril = order[tril]
                masked =  np.append(masked, tril)

            # Remove masked value to the iteration list
            order = np.delete(order, masked)
            # Get the indexes of nodes (i,j) for each observed interactions
            order = list(zip(*np.unravel_index(order, self.data_ma.shape)))
            self._order = order
        else:
            order = self._order

        if randomize is True:
            np.random.shuffle(order)
        return order

    # @debug: symmetric matrix ?
    def loglikelihood(self, j, i, k_ji, k_ij):
        w_ji = self.data_ma[j, i]
        self.word_topic_counts[w_ji, k_ji, k_ij] -= 1
        self.total_w_k[k_ji, k_ij] -= 1
        if self._symmetric:
            self.word_topic_counts[w_ji, k_ij, k_ji] -= 1
            self.total_w_k[k_ij, k_ji] -= 1

        if self.assortativity == 2:
            if k_ji == k_ij:
                log_smooth_count_ji = np.log(self.word_topic_counts[w_ji] + self.delta[w_ji])
                ll = log_smooth_count_ji - np.log(self.total_w_k + self.w_delta)
            else:
                ll = self.epsilon

        else:
            log_smooth_count_ji = np.log(self.word_topic_counts[w_ji] + self.delta[w_ji])
            ll = log_smooth_count_ji - np.log(self.total_w_k + self.w_delta)

        return ll


class ZSampler(object):
    # Alternative is to keep the two count matrix and
    # The docment-word topic array to trace get x(-ij) topic assignment:
        # C(word-topic)[w, k] === n_dotkw
        # C(document-topic[j, k] === n_jdotk
        # z DxN_k topic assignment matrix
        # --- From this reconstruct theta and phi

    def __init__(self, alpha_0, likelihood, K_init=1):
        self.K_init = K_init or 1
        self.alpha_0 = alpha_0

        self.likelihood = likelihood
        if self.likelihood._symmetric is not None:
            self.symmetric_pt = self.likelihood._symmetric +1 # the increment for Gibbs iteration
            self.nodes_list = likelihood.nodes_list
            self._nmap = likelihood._nmap
            self.data_dims = self.likelihood.data_dims
            self.J = len(self.data_dims)
            self.z = self._init_topics_assignement()
            self.doc_topic_counts = self.make_doc_topic_counts()
            if not hasattr(self, 'K'):
                # Nonparametric Case
                self.purge_empty_topics()
            self.likelihood.make_word_topic_counts(self.z, self.get_K())
            self.likelihood.total_w_k = self.likelihood.word_topic_counts.sum(0)

            # if a tracking of topics indexis pursuit,
            # pay attention to the topic added among those purged...(A topic cannot be added and purged in the same sample iteration !)
            self.last_purged_topics = []


    # @debug: symmetric matrix ?
    def _init_topics_assignement(self):
        dim = (self.J, self.J, 2)
        alpha_0 = self.alpha_0

        # Poisson way
        #z = np.array( [poisson(alpha_0, size=dim) for dim in data_dims] )

        # Random way
        K = self.K_init
        z = np.random.randint(0, K, (dim))

        if self.likelihood._symmetric:
            z[:, :, 0] = np.triu(z[:, :, 0]) + np.triu(z[:, :, 0], 1).T
            z[:, :, 1] = np.triu(z[:, :, 1]) + np.triu(z[:, :, 1], 1).T

        # LDA way
        # improve local optima ?
        #theta_j = dirichlet([1, gmma])
        #todo ?

        return z

    def sample(self):
        # Add pnew container
        self._update_log_alpha_beta()
        self.update_matrix_shape()

        lgg.debug('Sample z...')
        lgg.vdebug('#J \t #I \t  #topic')
        for j, i in self.likelihood.data_iter(randomize=True):
            lgg.vdebug( '%d \t %d \t %d' % ( j , i, self.doc_topic_counts.shape[1]-1))
            params = self.prob_zji(j, i, self._K + 1)
            sample_topic_raveled = categorical(params)
            k_j, k_i = np.unravel_index(sample_topic_raveled, (self._K+1, self._K+1))
            k_j, k_i = k_j[0], k_i[0] # beurk :(
            self.z[j, i, 0] = k_j
            self.z[j, i, 1] = k_i

            # Regularize matrices for new topic sampled
            if k_j == self.doc_topic_counts.shape[1] - 1 or k_i == self.doc_topic_counts.shape[1] - 1:
                self._K += 1
                #print 'Simplex probabilities: %s' % (params)
                self.update_matrix_shape(new_topic=True)

            self.update_matrix_count(j,i,k_j, k_i)

        # Remove pnew container
        self.doc_topic_counts = self.doc_topic_counts[:, :-1]
        self.likelihood.word_topic_counts = self.likelihood.word_topic_counts[:, :-1, :-1]
        self.purge_empty_topics()

        return self.z

    def update_matrix_shape(self, new_topic=False):
        if new_topic is True:
            # Updata alpha
            self.log_alpha_beta = np.append(self.log_alpha_beta, self.log_alpha_beta[-1])
            self.alpha = np.append(self.alpha, np.exp(self.log_alpha_beta[-1]))

        # Update Doc-topic count
        new_inst = np.zeros((self.J, 1), dtype=int)
        self.doc_topic_counts = np.hstack((self.doc_topic_counts, new_inst))

        # Update word-topic count
        new_feat_1 = np.zeros((self.likelihood.nfeat, self._K), dtype=int)
        new_feat_2 = np.zeros((self.likelihood.nfeat, self._K+1), dtype=int)
        self.likelihood.word_topic_counts = np.dstack((self.likelihood.word_topic_counts, new_feat_1))
        self.likelihood.word_topic_counts = np.hstack((self.likelihood.word_topic_counts, new_feat_2[:, None]))
        # sum all to update to fit the shape (a bit nasty if operation (new topic) occur a lot)
        self.likelihood.total_w_k = self.likelihood.word_topic_counts.sum(0)

    def update_matrix_count(self, j, i, k_j, k_i):
        self.doc_topic_counts[j, k_j] += self.symmetric_pt
        self.doc_topic_counts[i, k_i] += self.symmetric_pt
        self.likelihood.word_topic_counts[self.likelihood.data_ma[j,i], k_j, k_i] += 1
        self.likelihood.total_w_k[k_j, k_i] += 1
        if self.likelihood._symmetric:
            self.likelihood.word_topic_counts[self.likelihood.data_ma[j,i], k_i, k_j] += 1
            self.likelihood.total_w_k[k_i, k_j] += 1

    # @debug: symmetric matrix ?
    def make_doc_topic_counts(self):
        K = self.get_K()
        counts = np.zeros((self.J, K), dtype=int)

        for j, i in self.likelihood.data_iter(randomize=False):
            k_j = self.z[j, i, 0]
            k_i = self.z[j, i, 1]
            counts[j, k_j] += self.symmetric_pt
            counts[i, k_i] += self.symmetric_pt
        return counts

    def _update_log_alpha_beta(self):
        self.log_alpha_beta = np.log(self.alpha_0) + np.log(self.betasampler.beta)
        self.alpha = np.exp(self.log_alpha_beta)

    # Remove empty topic in nonparametric case
    # @debug: symmetric matrix ?
    def purge_empty_topics(self):
        counts = self.doc_topic_counts

        dummy_topics = []
        # Find empty topics
        for k, c in enumerate(counts.T):
            if c.sum() == 0:
                dummy_topics.append(k)
        for k in sorted(dummy_topics, reverse=True):
            counts = np.delete(counts, k, axis=1)
            self._K -= 1
            if hasattr(self.likelihood, 'word_topic_counts'):
                self.likelihood.word_topic_counts = np.delete(self.likelihood.word_topic_counts, k, axis=1)
                self.likelihood.word_topic_counts = np.delete(self.likelihood.word_topic_counts, k, axis=2)
            # Regularize Z
            for d in self.z:
                d[d > k] -= 1
            # Regularize alpha_beta, minus one the pnew topic
            if hasattr(self, 'log_alpha_beta') and k < len(self.log_alpha_beta)-1:
                self.log_alpha_beta = np.delete(self.log_alpha_beta, k)
                self.betasampler.beta = np.delete(self.betasampler.beta, k)

        self.last_purged_topics = dummy_topics
        if len(dummy_topics) > 0:
            lgg.debug( 'zsampler: %d topics purged' % (len(dummy_topics)))
        self.doc_topic_counts =  counts

    def add_beta_sampler(self, betasampler):
        self.betasampler = betasampler
        self._update_log_alpha_beta()

    def get_K(self):
        if not hasattr(self, '_K'):
            self._K =  np.max(self.z) + 1
        return self._K

    # Compute probabilityy to sample z_ij = k for each [K].
    # K is would be fix or +1 for nonparametric case
    def prob_zji(self, j, i, K):
        k_jji = self.z[j, i, 0]
        k_jij = self.z[j, i, 1]
        self.doc_topic_counts[j, k_jji] -= self.symmetric_pt
        self.doc_topic_counts[i, k_jij] -= self.symmetric_pt

        # Keep the outer product in memory
        p_jk = self.doc_topic_counts[j] + self.alpha
        p_ik = self.doc_topic_counts[i] + self.alpha
        outer_kk = np.outer(p_jk, p_ik)

        params = np.log(outer_kk) + self.likelihood.compute(j, i, k_jji, k_jij)
        params = params[:K, :K].ravel()
        return lognormalize(params)

    def get_log_alpha_beta(self, k):
        old_max = self.log_alpha_beta.shape[0]

        if k > (old_max - 1):
            return self.log_alpha_beta[old_max - 1]
        else:
            return self.log_alpha_beta[k]

    def predictive_topics(self, data):
        pass

    def estimate_latent_variables(self):
        # check if likelihood  is equal if removing dummy empty topics...
        if not hasattr(self, 'logalpha'):
            log_alpha_beta = self.log_alpha_beta
            new_k = self.get_K()+1 - len(log_alpha_beta)
            if new_k > 0:
                gmma = log_alpha_beta[-1]
                log_alpha_beta = np.hstack((log_alpha_beta, np.ones((new_k,))*gmma))
            # Remove empty possibly new topic
            alpha = np.exp(log_alpha_beta[:-1])
        else:
            alpha = np.exp(self.logalpha)
        delta = self.likelihood.delta
        K = len(alpha)

        # Recontruct Documents-Topic matrix
        _theta = self.doc_topic_counts + np.tile(alpha, (self.J, 1))
        self._theta = (_theta.T / _theta.sum(axis=1)).T

        # Recontruct Words-Topic matrix
        _phi = self.likelihood.word_topic_counts + np.tile(delta, (K, K, 1)).T
        #self._phi = (_phi / np.linalg.norm(_phi, axis=0))[1]
        self._phi = (_phi / _phi.sum(0))[1]

        return self._theta, self._phi

class ZSamplerParametric(ZSampler):
    # Parametric Version of HDP sampler. Number of topics fixed.

    def __init__(self, alpha_0, likelihood, K, alpha='asymmetric'):
        self.K = self.K_init = self._K =  K
        if 'alpha' in ('symmetric', 'fix'):
            alpha = np.ones(K)*1/K
        elif 'alpha' in ('asymmetric', 'auto'):
            alpha = np.asarray([1.0 / (i + np.sqrt(K)) for i in range(K)])
            alpha /= alpha.sum()
        else:
            alpha = np.ones(K)*alpha_0
        self.logalpha = np.log(alpha)
        self.alpha = alpha
        super(ZSamplerParametric, self).__init__(alpha_0, likelihood, self.K)

    def sample(self):
        lgg.debug('Sample z...')
        lgg.vdebug('#J \t #I \t #topic')
        for j, i in self.likelihood.data_iter(randomize=True):
            lgg.vdebug( '%d \t %d \t %d' % (j , i, self.doc_topic_counts.shape[1]-1))
            params = self.prob_zji(j, i, self.K)
            sample_topic_raveled = categorical(params)
            k_j, k_i = np.unravel_index(sample_topic_raveled, (self._K, self._K))
            k_j, k_i = k_j[0], k_i[0] # beurk :(
            self.z[j, i, 0] = k_j
            self.z[j, i, 1] = k_i
            nodes_classes_ass = [(j, k_j), (i, k_i)]

            self.update_matrix_count(j, i, k_j, k_i)
        return self.z

    def get_K(self):
        return self.K

    def get_log_alpha_beta(self, k):
        return self.logalpha[k]



class NP_CGS(object):

    # Joint Sampler of topic Assignement, table configuration, and beta proportion.
    # ref to direct assignement Sampling in HDP (Teh 2006)
    def __init__(self, zsampler, msampler, betasampler, hyper='auto', hyper_prior=None):
        zsampler.add_beta_sampler(betasampler)

        self.zsampler = zsampler
        self.msampler = msampler
        self.betasampler = betasampler

        msampler.sample()
        betasampler.sample()

        if hyper.startswith( 'auto' ):
            self.hyper = hyper
            if hyper_prior is None:
                self.a_alpha = 1
                self.b_alpha = 1
                self.a_gmma = 1
                self.b_gmma = 1
            else:
                self.a_alpha = hyper_prior[0]
                self.b_alpha = hyper_prior[1]
                self.a_gmma = hyper_prior[2]
                self.b_gmma = hyper_prior[3]

            self.optimize_hyper_hdp()
        elif hyper.startswith( 'fix' ):
            self.hyper = hyper
        else:
            raise NotImplementedError('Hyperparameters optimization ?')

    def optimize_hyper_hdp(self):
        # Optimize \alpha_0
        m_dot = self.msampler.m_dotk.sum()
        alpha_0 = self.zsampler.alpha_0
        n_jdot = np.array(self.zsampler.data_dims, dtype=float) # @debug add row count + line count for masked !
        #p = np.power(n_jdot / alpha_0, np.arange(n_jdot.shape[0]))
        #norm = np.linalg.norm(p)
        #u_j = binomial(1, p/norm)
        u_j = binomial(1, alpha_0/(n_jdot + alpha_0))
        #u_j = binomial(1, n_jdot/(n_jdot + alpha_0))
        try:
            v_j = beta(alpha_0 + 1, n_jdot)
        except:
             #n_jdot[n_jdot == 0] = np.finfo(float).eps
            lgg.warning('Unable to optimize MMSB parameters, possible empty sequence...')
            return
        shape_a = self.a_alpha + m_dot - u_j.sum()
        if shape_a <= 0:
            lgg.warning('Unable to optimize MMSB parameters, possible empty sequence...')
            return
        new_alpha0 = gamma(shape_a, 1/( self.b_alpha - np.log(v_j).sum()), size=3).mean()
        self.zsampler.alpha_0 = new_alpha0

        # Optimize \gamma
        K = self.zsampler._K
        #m_dot = self.msampler.m_dotk
        gmma = self.betasampler.gmma
        #p = np.power(m_dot / gmma, np.arange(m_dot.shape[0]))
        #norm = np.linalg.norm(p)
        #u = binomial(1, p/norm)
        u = binomial(1, gmma / (m_dot + gmma))
        #u = binomial(1, m_dot / (m_dot + gmma))
        v = beta(gmma + 1, m_dot)
        new_gmma = gamma(self.a_gmma + K -1 + u, 1/(self.b_gmma - np.log(v)), size=3).mean()
        self.betasampler.gmma = new_gmma

        #print 'm_dot %d, alpha a, b: %s, %s ' % (m_dot, self.a_alpha + m_dot - u_j.sum(), 1/( self.b_alpha - np.log(v_j).sum()))
        #print 'gamma a, b: %s, %s ' % (self.a_gmma + K -1 + u, 1/(self.b_gmma - np.log(v)))
        lgg.debug('hyper sample: alpha_0: %s gamma: %s' % (new_alpha0, new_gmma))
        return

    def sample(self):
            z = self.zsampler.sample()
            m = self.msampler.sample()
            beta = self.betasampler.sample()

            if self.hyper.startswith('auto'):
                self.optimize_hyper_hdp()

            return z, m, beta

class CGS(object):

    def __init__(self, zsampler):
        self.zsampler = zsampler

    def sample(self):
        return self.zsampler.sample()


class GibbsRun(GibbsSampler):
    __abstractmethods__ = 'model'
    def __init__(self, expe, frontend):
        self.comm = dict() # Empty dict to store communities and blockmodel structure
        self._csv_typo = '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'
        #self._fmt = '%d %.4f %.8f %.8f %d %.8f %.8f %.4f %.4f %.4f %.4f'
        GibbsSampler.__init__(self, expe, frontend)

    def limit_k(self, N, directed=True):
        alpha, gmma, delta = self.get_hyper()
        N = int(N)
        if directed is True:
            m = alpha * N * (digamma(N+alpha) - digamma(alpha))
        else:
            m = alpha * N * (digamma(2*N+alpha) - digamma(alpha))

        # Number of class in the CRF
        K = int(gmma * (digamma(m+gmma) - digamma(gmma)))
        return K

    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        if mode == 'generative' :
            self.update_hyper(hyperparams)
            alpha, gmma, delta = self.get_hyper()
            N = int(N)
            _name = self.__module__.split('.')[-1]
            if _name == 'immsb_cgs':
                # @todo: compute the variance for random simulation
                # Number of table in the CRF
                if symmetric is True:
                    m = alpha * N * (digamma(N+alpha) - digamma(alpha))
                else:
                    m = alpha * N * (digamma(2*N+alpha) - digamma(alpha))

                # Number of class in the CRF
                K = int(gmma * (digamma(m+gmma) - digamma(gmma)))
                alpha = gem(gmma, K)

            i = 0
            while i<3:
                try:
                    dirichlet(alpha, size=N)
                    i=0
                    break
                except ZeroDivisionError:
                    # Sometimes umprobable values !
                    alpha = gem(gmma, K)
                    i += 1

            # Generate Theta
            if i > 0:
                params, order = zip(*np.sorted(zip(alpha, range(len(alpha)), reverse=True)))
                _K = int(1/3. * len(alpha))
                alpha[order[:_K]] = 1
                alpha[order[_K:]] = 0
                theta = multinomial(1, alpha, size=N)
            else:
                theta = dirichlet(alpha, size=N)

            # Generate Phi
            phi = beta(delta[0], delta[1], size=(K,K))
            if symmetric is True:
                phi = np.triu(phi) + np.triu(phi, 1).T

            self._theta = theta
            self._phi = phi
        elif mode == 'predictive':
            try: theta, phi = self.get_params()
            except: return self.generate(N, K, hyperparams, 'generative', symmetric)
            K = theta.shape[1]

        pij = self.likelihood(theta, phi)

        # Treshold
        #pij[pij >= 0.5 ] = 1
        #pij[pij < 0.5 ] = 0
        #Y = pij

        # Sampling
        pij = np.clip(pij, 0, 1)
        Y = sp.stats.bernoulli.rvs(pij)

        #for j in xrange(N):
        #    print 'j %d' % j
        #    for i in xrange(N):
        #        zj = categorical(theta[j])
        #        zi = categorical(theta[i])
        #        Y[j, i] = sp.stats.bernoulli.rvs(B[zj, zi])
        return Y, theta, phi


    def get_clusters(self, K=None, skip=0):
        """ Return a vector of clusters membership of nodes.

        Parameters
        ----------
        K : int
          Number of clusters. if None, used the number learn at inference.
        below: int
          skip the x firt bigger class.
        """
        theta, phi = self.get_params()
        clusters = np.argmax(theta.dot(phi), axis=1)
        if not K: return clusters
        hist = np.bincount(clusters)
        sorted_clusters = sorted(zip(hist, range(len(hist))), reverse=True)[skip:K+skip]
        _, strong_c = zip(*sorted_clusters)
        mask = np.ones(theta.shape)
        mask[:, strong_c] = 0
        theta_ma = ma.array(theta, mask=mask)
        clusters = np.argmax(theta_ma, axis=1)
        return clusters

    #add sim optionsin clusters
    def get_communities(self, K=None, skip=0):
        """ Return a vector of clusters membership of nodes.

        Parameters
        ----------
        K : int
          Number of clusters. if None, used the number learn at inference.
        """
        theta, phi = self.get_params()
        K = K or theta.shape[1]

        # Kmeans on Similarity
        _phi = phi.copy()
        _phi[phi < phi.mean()] = 0
        clusters = kmeans(theta.dot(phi), K=K)
        #s = sorted(zip(phi.ravel(), np.arange(phi.size)), reverse=True)[:K]
        #strong_c  = np.unravel_index(zip(*s)[1])
        #sim = theta.dot(phi).dot(theta.T)

        # Strongest communities
        #hist = phi.diagonal()
        #sorted_clusters = sorted(zip(hist, range(len(hist))), reverse=True)[skip:K+skip]
        #_, strong_c = zip(*sorted_clusters)
        #mask = np.ones(theta.shape)
        #mask[:, strong_c] = 0
        #theta_ma = ma.array(theta, mask=mask)
        #clusters = np.argmax(theta_ma, axis=1)

        return clusters

    #@wrapper !
    ### Degree by class (maxassignment)
    def communities_analysis(self, data, clustering='modularity'):
        symmetric = (data == data.T).all()

        if hasattr(self, clustering):
            f = getattr(self, clustering)
        else:
            raise NotImplementedError('Clustering algorithm unknow')

        if not hasattr(self, 'comm'):
            self.comm = dict()

        self.comm.update( f(data=data, symmetric=symmetric) )
        return self.comm

    def max_assignement(self, **kwargs):
        data = kwargs['data']
        symmetric = kwargs['symmetric']

        clusters = self.get_clusters()
        block_hist = np.bincount(clusters)

        local_degree = {}
        for n, c in enumerate(clusters):
            comm = str(c)
            local = local_degree.get(comm, [])
            degree_n = data[n,:].sum()
            #degree_n = data[n,:][clusters == c].sum()
            if not symmetric:
                degree_n += data[:, n].sum()
                #degree_n += data[:, n][clusters == c].sum()
            local.append(degree_n)
            local_degree[comm] = local

        return {'local_degree':local_degree,
                'clusters':clusters,
                'block_hist': block_hist,
                'size': len(block_hist)}

    ### Degree by class (sitll max assigment !)
    def modularity(self, **kwargs):
        data = kwargs['data']
        symmetric = kwargs['symmetric']

        clusters = self.get_clusters()
        block_hist = np.bincount(clusters)

        local_degree = {}
        if symmetric:
            k_perm = np.unique( list(map(list, map(set, itertools.product(np.unique(clusters) , repeat=2)))))
        else:
            k_perm = itertools.product(np.unique(clusters) , repeat=2)

        for c in k_perm:
            if type(c) in (np.float64, np.int64):
                # one clusters (as it appears for real with max assignment
                l = k = c
            elif  len(c) == 2:
                # Stochastic Equivalence (extra class bind
                k, l = c
            else:
                # Comunnities (intra class bind)
                k = l = c.pop()
            comm = '-'.join([str(k), str(l)])
            local = local_degree.get(comm, [])

            C = np.tile(clusters, (data.shape[0],1))
            y_c = data * ((C==k) & (C.T==l))
            if y_c.size > 0:
                local_degree[comm] = list(adj_to_degree(y_c).values())

            # Summing False !
            #for n in np.arange(data.shape[0]))[clusters == k]:
            #    degree_n = data[n,:][(clusters == k) == (clusters == l)].sum()
            #    if not symmetric:
            #        degree_n = data[n,:][(clusters == k) == (clusters == l)].sum()
            #    local.append(degree_n)
            #local_degree[comm] = local

        return {'local_degree':local_degree,
                'clusters':clusters,
                'block_hist': block_hist,
                'size': len(block_hist)}

    def blockmodel_ties(self, data, remove_empty=True):
        """ return ties based on a weighted average
            of the local degree ditribution """

        if not hasattr(self, 'comm'):
            self.communities_analysis(data)
        comm = self.comm

        m = [ wmean(a[0], a[1], mean='arithmetic')  for d in comm['local_degree'].values() for a in [degree_hist(d)] if len(a[0])>0 ]
        # Mean
        #m = map(np.mean, comm['local_degree'].values())
        # Variance Unused How to represant that !?
        #v = np.array(map(np.std, comm['local_degree'].values()))

        # factorize by using clusters hist instead
        hist, label = sorted_perm(m, comm['local_degree'].keys(), reverse=True)

        if remove_empty is True:
            null_classes = (hist == 0).sum()
            if null_classes > 0:
                hist = hist[:-null_classes]; label = label[:-null_classes]

        bm = zip(label, hist)
        self.comm['block_ties'] = bm
        return bm

    def compute_measures(self, begin_it=0):
        self._K = self.s.zsampler._K

        self._entropy = self.compute_entropy()
        #self._entropy_t = self.predictive_likelihood()
        self._entropy_t = np.nan

        alpha_0 = self.s.zsampler.alpha_0
        try:
            gmma = self.s.betasampler.gmma
            alpha = np.exp(self.s.zsampler.log_alpha_beta)
        except:
            gmma = np.nan
            alpha = np.exp(self.s.zsampler.logalpha)

        self._alpha = alpha_0
        self._gmma = gmma
        self.alpha_mean = alpha.mean()
        self.alpha_var = alpha.var()
        self.delta_mean = self.s.zsampler.likelihood.delta.mean()
        self.delta_var = self.s.zsampler.likelihood.delta.var()

        self.time_it = time() - begin_it

    # Mean can be arithmetic or geometric
    def compute_entropy(self):
        buritos = self.s.zsampler
        self._theta, self._phi = buritos.estimate_latent_variables()
        pij = self.likelihood(self._theta, self._phi)

        # Log-likelihood
        pij = buritos.likelihood.data_A * pij + buritos.likelihood.data_B
        ll = np.log(pij).sum()

        # Entropy
        entropy = ll
        #entropy = - ll / buritos.likelihood.nnz

        # Perplexity is 2**H(X).

        return entropy

    def compute_entropy_t(self):
        buritos = self.s.zsampler
        self._theta, self._phi = buritos.estimate_latent_variables()
        pij = self.likelihood(self._theta, self._phi)

        # Log-likelihood
        pij = buritos.likelihood.data_A_t * pij + buritos.likelihood.data_B_t
        ll = np.log(pij).sum()

        # Entropy
        entropy = ll
        #entropy = - ll / buritos.likelihood.nnz

        # Perplexity is 2**H(X).

        return entropy

    def purge(self):
        self.s.zsampler.betasampler = None
        self.s.zsampler._nmap = None
        self.s.msampler = None
        self.s.betasampler = None
        self.s.zsampler.likelihood = None

    def get_hyper(self):
        if not hasattr(self, '_alpha'):
            try:
                self._delta = self.s.zsampler.likelihood.delta
                if type(self.s) is NP_CGS:
                    self._alpha = self.s.zsampler.alpha_0
                    self._gmma = self.s.betasampler.gmma
                else:
                    self._alpha = self.s.zsampler.alpha
                    self._gmma = None
            except:
                lgg.error('Need to propagate hyperparameters to BaseModel class')
                self._delta = None
                self._alpha = None
                self._gmma =  None
        return self._alpha, self._gmma, self._delta


    def sample(self):
        self.s.sample()



import sys, os, re
import pickle, json
import pkgutil, pyclbr, inspect, ast
from importlib import import_module
from collections import OrderedDict
from itertools import groupby
import zlib
import numpy as np

from pymake import ExpDesign, ExpeFormat, get_pymake_settings

import logging
lgg = logging.getLogger('root')




''' Python I/O Parsing Tools. '''



class PyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(PyEncoder, self).default(obj)


def resolve_filename(fn, ext='pk'):
    splt = fn.split('.')
    if splt[-1] not in ['pk', 'json', 'gt']:
        fn += '.' + ext
    return fn

def load(fn, ext='pk', silent=False, driver=None):
    fn = resolve_filename(fn, ext)
    ext = fn.split('.')[-1]

    # Seek for compressed data
    if os.path.exists(fn+'.gz'):
        compressed = True
        fn += '.gz'
    else:
        compressed = False
        lgg.debug('Data is not compressed: %s' % fn)

    if not silent:
        lgg.info('Loading data : %s' % fn)

    if driver:
        return driver(fn)
    elif ext == 'pk':
        if compressed:
            with open(fn, 'rb') as _f:
                return pickle.loads(zlib.decompress(_f.read()))
        else:
            with open(fn, 'rb') as _f:
                return pickle.load(_f)
        #except:
        #    # python 2to3 bug
        #    _f.seek(0)
        #    try:
        #        model =  pickle.load(_f, encoding='latin1')
        #    except OSError as e:
        #        cls.log.critical("Unknonw error while opening  while `_load_model' at file: %s" % (fn))
        #        cls.log.error(e)
        #            return
    elif ext == 'json':
        if compressed:
            with open(fn, 'rb') as _f:
                return json.loads(zlib.decompress(_f.read()))
        else:
            with open(fn) as _f:
                return json.load(_f)
    else:
        raise ValueError('File format unknown: %s' % ext)


def save(fn, data, ext='pk', silent=False, compress=None, driver=None,
         compressed_pk=True, compressed_json=False):

    if compress is not None:
        compressed_pk = compressed_json = compressed

    fn = resolve_filename(fn, ext)
    ext = fn.split('.')[-1]

    if not silent:
        lgg.info('Saving data : %s' % fn)

    if driver:
        if data is None:
            return driver(fn)
        else:
            return driver(fn, data)
    elif ext == 'pk':
        if compressed_pk:
            fn += '.gz'
            obj = zlib.compress(pickle.dumps(data))
            with open(fn, 'wb') as _f:
                return _f.write(obj)
        else:
            with open(fn, 'wb') as _f:
                return pickle.dump(data, _f, protocol=pickle.HIGHEST_PROTOCOL)

    elif ext == 'json':
        # @Todo: Option to update if file exists:
        #res = json.load(open(fn,'r'))
        #res.update(data)
        #self.log.info('Updating json data: %s' % fn)
        if compressed_json:
            fn += '.gz'
            obj = zlib.compress(json.dumps(data))
            with open(fn, 'wb') as _f:
                return _f.write(obj)
        else:
            with open(fn, 'wb') as _f:
                return json.dump(data, _f, cls=PyEncoder)
    else:
        raise ValueError('File format unknown: %s' % ext)




def is_abstract(cls):
    ABCFLAG = '__abstractmethods__'
    isabc = hasattr(cls, ABCFLAG) or inspect.isabstract(cls)
    return isabc

def is_empty_file(filen):

    if os.path.exists(filen+'.gz'):
        filen = filen+'.gz'

    if not os.path.isfile(filen) or os.stat(filen).st_size == 0:
        return True

    try:
        with open(filen, 'r') as f: first_line = f.readline()
    except UnicodeDecodeError:
        first_line = '...'

    if first_line[0] in ('#', '%') and sum(1 for line in open(filen)) <= 1:
        # empy file
        return True
    else:
       return False

# untested!
def get_decorators(cls):
    target = cls
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = []
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[node.name].append(name)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(inspect.getsource(target)))
    return decorators


class PackageWalker(object):
    ''' Import all submodules of a module, recursively.

        Module walker implements this class by overloading the methods 'submodule_hook',
        that should call one of the following methods :
            * submodul_hook_by_attr : to reach/get module if an attribut is {defined}
            * submodul_hook_by_class : to reach/get a module if a class is of a {defined} type.

        The {defined} value should be defined in the {_default_attr} with a class inherit PackageWalker is defining the class.
        (see the *Loader for .e.g).

        Attributes
        ----------
        packages : OrederedDict
            contains the module by name
        _cls_browse : pyclbr
            contains module informations
    '''

    _default_attr = dict(
        attr_filter = None,
        class_filter = None,
        max_depth = 1,
        prefix = False,
        shrink_module_name = True,
    )

    def __init__(self, module_name, **kwargs):
        [setattr(self, k, kwargs.get(k, v)) for k,v in self._default_attr.items()]

        self._cls_browse = {}
        self._unavailable_modules = []

        ### Core function
        self.packages = self._get_packages(module_name)

        if len(self._unavailable_modules) > 0:
            lgg.warning('There is some unavailable modules : %s' % self._unavailable_modules)

    def __repr__(self):
        return self.packages

    def _get_packages(self, module_name,  _depth=0):
        ''' recursive call with regards to depth parameter '''

        if self.prefix is True:
            prefix = module_name
        elif self.prefix is False:
            prefix = ''
        elif isinstance(self.prefix, str):
            prefix = self.prefix

        packages = OrderedDict()
        try:
            module = import_module(module_name)
        except ImportError as e:
            lgg.warning('package unavailable (%s) : %s' % (e, module_name))
            return packages

        for loader, name, is_pkg in pkgutil.walk_packages(module.__path__):
            submodule_name = module_name + '.' + name
            if is_pkg and _depth < self.max_depth:
                next_depth = _depth + 1
                packages.update(self._get_packages(submodule_name, next_depth))
                continue
            try:
                submodule = import_module(submodule_name)
            except ImportError as e:
                lgg.debug('submodule unavailable (%s) : %s'%(e, submodule_name))
                self._unavailable_modules.append(submodule_name)
                if 'algo' in str(e):
                    lgg.warning('Please report this error: pymake.io algo condition ?')
                    submodule = import_module(submodule_name)
                continue
            except Exception as e:
                lgg.critical('Module Error: %s' % e)
                continue

            spck = self.submodule_hook(submodule, prefix, name=name)
            if spck:
                packages.update(spck)

        return packages

    def submodule_hook(self, *args, **kwargs):
        raise NotImplementedError

    def submodule_hook_by_attr(self, submodule, prefix, name=None):
        try:
            _packages = pyclbr.readmodule(submodule.__name__)
        except:
            # @DEBUG : use shelve to get it
            sys.path.append(os.getenv('PWD'))
            #_packages = pyclbr.readmodule(submodule.__name__)
            _packages = pyclbr.readmodule('.'.join(submodule.__name__.split('.')[1:]))
        self._cls_browse.update(_packages)
        packages = {}
        for cls_name, cls  in _packages.items():
            obj = getattr(submodule, cls_name)
            if not hasattr(obj, self.attr_filter) or is_abstract(obj):
            #if not any([hasattr(obj, attr) for attr in self.attr_filter]) or is_abstract(obj):
                continue
            if self.shrink_module_name:
                prefix = prefix.split('.')[0]
            if prefix:
                name = prefix +'.'+ cls_name.lower()
            else:
                name = cls_name.lower()

            packages[name] = obj

        return packages

    def submodule_hook_by_class(self, submodule, prefix, name=None):
        try:
            _packages = pyclbr.readmodule(submodule.__name__)
        except:
            # @DEBUG : use shelve to get it
            sys.path.append(os.getenv('PWD'))
            #_packages = pyclbr.readmodule(submodule.__name__)
            _packages = pyclbr.readmodule('.'.join(submodule.__name__.split('.')[1:]))
        self._cls_browse.update(_packages)
        packages = {}
        for cls_name, cls  in _packages.items():
            obj = getattr(submodule, cls_name)
            if not issubclass(obj, self.class_filter) or is_abstract(obj):
            # ExpDesign don't load without type() ?
            #if not issubclass(type(obj), self.class_filter) or is_abstract(obj):
                continue

            if prefix:
                name = prefix +'.'+ cls_name.lower()
            else:
                name = cls_name.lower()
            packages[name] = obj

        return packages

class ModelsLoader(PackageWalker):
    ''' Load models in a modules :
        * lookup for models inside a module by fuzzing the name of the module.
        * Need either a :
            * `fit' method to be identified as model,
            * a `module' attribute name => that should have a fit method.
    '''
    def submodule_hook(self, *args,  **kwargs):
        return super(ModelsLoader, self).submodule_hook_by_attr(*args,  **kwargs)

    @classmethod
    def get_packages(cls, module_name, **kwargs):
        if not 'attr_filter' in kwargs:
            kwargs['attr_filter'] = 'fit'

        # need of not needs ?
        #if isinstance(module_name, list):
        #    packs = {}
        #    for m in module_name:
        #        packs.update(cls(m, **kwargs).packages)
        #    return packs
        #else:
        #    return cls(module_name, **kwargs).packages

        return cls(module_name, **kwargs).packages

    @staticmethod
    def _fuzz_funcname(name):
        ''' Fuzzing search '''
        # use regexp ?
        # @Obsolete: use attr_method only...
        nsplit = name.split('_')
        nsplit1 = ''.join([nsplit[0].upper()] + nsplit[1:])
        nsplit2 = ''.join(list(map(str.title, nsplit)))
        return (name, name.lower(), name.upper(), name.title(),
                name.lower().title(), nsplit1, nsplit2)

    def _submodule_hook_by_fuzz(self, submodule, prefix, name):
        obj_list = dir(submodule)
        idx = None
        packages = OrderedDict()
        for n in self._fuzz_funcname(name):
            # Search for the object inside the submodule
            try:
                idx = obj_list.index(n)
                break
            except:
                continue
        if idx is None: return

        obj = getattr(submodule, obj_list[idx])
        if hasattr(obj, self.attr_filter):
            if prefix:
                name = prefix +'.'+ name
            # remove duplicate name
            name = '.'.join([x[0] for x in groupby(name.split('.'))])
            packages[name] = obj

        return packages

    @classmethod
    def get_atoms(cls,  _type='short'):
        if _type == 'short':
            shrink_module_name = True
        elif _type == 'topos':
            shrink_module_name = False

        packages = get_pymake_settings('_model')
        atoms = OrderedDict()
        for pkg in packages:
            if len(pkg) > 8:
                prefix = pkg[:3]
                if '.' in pkg:
                    prefix  += ''.join(map(lambda x:x[0], pkg.split('.')[1:]))
            else:
                prefix = True
            atoms.update(ModelsLoader.get_packages(pkg,  prefix=prefix, max_depth=3, shrink_module_name=shrink_module_name))
        return atoms

class CorpusLoader(PackageWalker):
    def submodule_hook(self, *args, **kwargs):
        raise NotImplemented('todo :scikit-learn loohup')

    #@classmethod
    #def get_atoms(cls_type=None):
    #    # get some information about what package to use in _spec ...
    #    atoms = cls.get_packages('pymake.data')
    #    return atoms

class ScriptsLoader(PackageWalker):

    module = ExpeFormat

    def submodule_hook(self, *args,  **kwargs):
        return super(ScriptsLoader, self).submodule_hook_by_class(*args, **kwargs)

    @classmethod
    def get_packages(cls, **kwargs):
        module_name = get_pymake_settings('_script')
        if not 'class_filter' in kwargs:
            kwargs['class_filter'] = cls.module

        if isinstance(module_name, list):
            packs = {}
            for m in module_name:
                packs.update(cls(m, **kwargs).packages)
            return packs
        else:
            return cls(module_name, **kwargs).packages

    @classmethod
    def get_atoms(cls):
        atoms = dict()
        modules = get_pymake_settings('_script')
        modules = [modules] if type(modules) is str else modules
        for module in modules:

            s = cls(module, class_filter=cls.module)

            ## get decorator for each class
            #class2met2dec = {}
            #for method, _class in classs.packages.items():
            #    append decoratpr information to filter @atpymake

            for surname, _module in s.packages.items():
                name = _module.__name__
                module = s._cls_browse[name]
                methods = list(module.methods.keys())
                for m in methods.copy():
                    _m = getattr(s.packages[name.lower()], m)
                    if not inspect.isfunction(_m) and m != '__call__':
                        methods.remove(m)
                    elif '__call__' == m:
                        methods.remove('__call__')
                        methods.append(name.lower())
                    elif m.startswith('_'):
                        methods.remove(m)
                    elif m in dir(cls.module):
                        methods.remove(m)

                content = {}
                content['scriptname'] = name
                content['scriptsurname'] = surname
                content['module_file'] = module.file
                content['module'] = _module.__module__
                content['_module'] = _module
                #content['module_name'] = '.'.join((module.name, module.module))
                content['module_super'] = module.super
                content['methods'] = set(methods)
                atoms[name] = content

        return atoms

class SpecLoader(PackageWalker):
    ''' Load specification of design experimentation :
        * Lookup for ExpDesign subclass in the target module.
    '''

    module = ExpDesign

    def submodule_hook(self, *args, **kwargs):
        return super(SpecLoader, self).submodule_hook_by_class(*args, **kwargs)

    @staticmethod
    def default_spec():
        import pymake
        return getattr(pymake, '__spec')

    #@staticmethod
    #def _default_spec():
    #    #module_name = get_pymake_settings('_spec')
    #    module_name = 'pymake.spec.netw'
    #    spec = import_module(module_name)
    #    for m in dir(spec):
    #        o = getattr(spec,m)
    #        if issubclass(o, cls.module) and not o is cls.module:
    #            return o()

    @classmethod
    def get_packages(cls, **kwargs):
        module_name = get_pymake_settings('_spec')
        if not 'class_filter' in kwargs:
            kwargs['class_filter'] = cls.module

        if isinstance(module_name, list):
            packs = {}
            for m in module_name:
                packs.update(cls(m, **kwargs).packages)
            return packs
        else:
            return cls(module_name, **kwargs).packages

    @classmethod
    def get_atoms(cls):
        expe_designs = []
        atoms = dict()

        modules = get_pymake_settings('_spec')
        modules = [modules] if type(modules) is str else modules
        for module in modules:

            s = cls(module, class_filter=cls.module)

            for surname, _module in s.packages.items():
                name = _module.__name__
                module = s._cls_browse[name]

                expd = getattr(import_module(module.module), name)()

                content = {}
                content['script_name'] = surname
                content['module_name'] = '.'.join((_module.__module__, _module.__name__))
                content['_module'] = module
                content['exp'] = expd._specs()
                atoms[name] = content

        return atoms














#
# @Obsolete Code:
#   Altough, there is some idea to re-build spec/paramas
#   by path traversing. It coul also be used to discover expe
#   from path traversing. But re-implementing it is advised.
#
#


### directory/file tree reference
# Default and New values
# @filename to rebuild file
_MASTERKEYS = OrderedDict((
    ('corpus'      , None),
    ('repeat'      , None),
    ('model'       , None),
    ('K'           , None),
    ('hyper'       , None),
    ('homo'        , None),
    ('N'           , 'all'),
))


# Debug put K_end inside the json
#_Key_measures = [ 'g_precision', 'Precision', 'Recall', 'K',
#                 'homo_dot_o', 'homo_dot_e', 'homo_model_o', 'homo_model_e',
#                ]
_Key_measures = [ 'g_precision', 'Precision', 'Recall', 'K',
                 'homo_dot_o', 'homo_dot_e', 'homo_model_o', 'homo_model_e',
                 'f1'
                ]

_New_Dims = [{'measure':len(_Key_measures)}]






def tree_hook(key, value):
    hook = False
    if key == 'corpus':
        if value in ('generator', ):
            hook = True
    return hook


# Warning to split('_')
def get_conf_from_file(target, mp):
    """ Return dictionary of property for an expe file.
        @mp: map parameters
        format model_K_hyper_N
        @template_file order important to align the dictionnary.
        """
    masterkeys = _MASTERKEYS.copy()
    template_file = masterkeys.keys()
    ##template_file = 'networks/generator/Graph13/debug11/immsb_10_auto_0_all.*'

    data_path = get_pymake_settings('project_data')
    # Relative path ignore
    if target.startswith(data_path):
        target.replace(data_path, '')

    path = target.lstrip('/').split('/')

    _prop = os.path.splitext(path.pop())[0]
    _prop = path + _prop.split('_')

    prop = {}
    cpt_hook_master = 0
    cpt_hook_user = 0
    # @Debug/Improve the nasty Hook here
    def update_pt(cur, master, user):
        return cur - master + user

    #prop = {k: _prop[i] for i, k in enumerate(template_file) if k in mp}
    for i, k in enumerate(template_file):
        if not k in mp:
            cpt_hook_master += 1
            continue
        pt = update_pt(i, cpt_hook_master, cpt_hook_user)
        hook = tree_hook(k, _prop[pt])
        if hook:
            cpt_hook_user += 1
            pt = update_pt(i, cpt_hook_master, cpt_hook_user)
        prop[k] = _prop[pt]

    return prop

def get_conf_dim_from_files(targets, mp):
    """ Return the sizes of proporties in a list for expe files
        @mp: map parameters """
    c = []
    for t in targets:
        c.append(get_conf_from_file(t, mp))

    sets = {}
    keys_name = mp.keys()
    for p in keys_name:
        sets[p] = len(set([ _p[p] for _p in c ]))

    return sets

def forest_tensor(target_files, map_parameters):
    """ It has to be ordered the same way than the file properties.
        Fuze directory to find available files then construct the tensor
        according the set space fomed by object found.
        @in target_files has to be orderedDict to align the the tensor access.
    """
    # Expe analyser / Tabulyze It

    # res shape ([expe], [model], [measure]
    # =================================================================================
    # Expe: [debug, corpus] -- from the dirname
    # Model: [name, K, hyper, homo] -- from the expe filename
    # measure:
    #   * 0: global precision,
    #   * 1: local precision,
    #   * 2: recall

    ### Output: rez.shape rez_map_l rez_map
    if not target_files:
        lgg.info('Target Files empty')
        return None

    #dim = get_conf_dim_from_files(target_files, map_parameters) # Rely on Expe...
    dim = dict( (k, len(v)) if isinstance(v, (list, tuple)) else (k, len([v])) for k, v in map_parameters.items() )

    rez_map = map_parameters.keys() # order !
    # Expert knowledge value
    new_dims = _New_Dims
    # Update Mapping
    [dim.update(d) for d in new_dims]
    [rez_map.append(n.keys()[0]) for n in new_dims]

    # Create the shape of the Ananisys/Resulst Tensor
    #rez_map = dict(zip(rez_map_l, range(len(rez_map_l))))
    shape = []
    for n in rez_map:
        shape.append(dim[n])

    # Create the numpy array to store all experience values, whith various setings
    rez = np.zeros(shape) * np.nan

    not_finished = []
    info_file = []
    for _f in target_files:
        prop = get_conf_from_file(_f, map_parameters)
        pt = np.empty(rez.ndim)

        assert(len(pt) - len(new_dims) == len(prop))
        for k, v in prop.items():
            try:
                v = int(v)
            except:
                pass
            try:
                idx = map_parameters[k].index(v)
            except Exception as e:
                lgg.error(prop)
                lgg.error('key:value error --  %s, %s'% (k, v))
                raise ValueError
            pt[rez_map.index(k)] = idx

        f = os.path.join(get_pymake_settings('project_data'), _f)
        d = load(f)
        if not d:
            not_finished.append( '%s not finish...\n' % _f)
            continue

        try:
            pt = list(pt.astype(int))
            for i, v in enumerate(_Key_measures):
                pt[-1] = i
                ### HOOK
                # v:  is the measure name
                # json_v: the value of the measure
                if v == 'homo_model_e':
                    try:
                        json_v =  d.get('homo_model_o') - d.get(v)
                    except: pass
                elif v == 'f1':
                    precision = d.get('Precision')
                    try:
                        recall = d.get('Recall')
                        recall*2
                    except:
                        # future remove
                        recall = d.get('Rappel')
                    json_v = 2*precision*recall / (precision+recall)
                else:
                    if v == 'Recall':
                        try:
                            v * 2
                        except:
                            v = 'Rappel'

                    json_v = d.get(v)
                rez[zip(pt)] = json_v

        except IndexError as e:
            lgg.error(e)
            lgg.error('Index Error: Files are probably missing here to complete the results...\n')

        #info_file.append( '%s %s; \t K=%s\n' % (corpus_type, f, K) )

    lgg.debug(''.join(not_finished))
    #lgg.debug(''.join(info_file))
    rez = np.ma.masked_array(rez, np.isnan(rez))
    return rez

def clean_extra_expe(expe, map_parameters):
    for k in expe:
        if k not in map_parameters and k not in [ k for d in _New_Dims for k in d.keys() ] :
            del expe[k]
    return expe

def make_tensor_expe_index(expe, map_parameters):
    ptx = []
    expe = clean_extra_expe(expe, map_parameters)
    for i, o in enumerate(expe.items()):
        k, v = o[0], o[1]
        if v in ( '*',): #wildcar / indexing ...
            ptx.append(slice(None))
        elif k in map_parameters:
            ptx.append(map_parameters[k].index(v))
        elif type(v) is int:
            ptx.append(v)
        elif type(v) is str and ':' in v: #wildcar / indexing ...
            sls = v.split(':')
            sls = [None if not u else int(u) for u in sls]
            ptx.append(slice(*sls))
        else:
            raise ValueError('Unknow data type for tensor forest')

    ptx = tuple(ptx)
    return ptx




import os
import numpy as np
import scipy as sp
import scipy.sparse as sparse
from scipy.sparse import lil_matrix, csr_matrix

from .frontend import DataBase
from .drivers import OnlineDatasetDriver

try:
    import graph_tool as gt
    from graph_tool import stats, clustering, inference, spectral, topology, generation, search
    from graph_tool import draw
except Exception as e:
    print('Error while importing graph-tool: %s' % e)


class frontendNetwork_gt(DataBase, OnlineDatasetDriver):

    ''' A frontend layer for the graph-tool object.

        Attributes
        ----------
        data: Graph (from gt)

    '''

    def __init__(self, expe, data, corpus=None):
        super().__init__(expe)
        self._data_type = 'network'
        corpus = corpus or {}

        self.data = data

        force_directed = expe.get('directed', corpus.get('directed'))
        force_directed = bool(force_directed) if force_directed is not None else None
        remove_self_loops = expe.get('remove_self_loops', True)

        # Remove selfloop
        if remove_self_loops:
            # @Warning: the corresponding weight are kept in the map properties
            self.log.debug('Self-loop are assumed to be remove from the graph.')
            self.remove_self_loops()

        # Set the direction and weights
        if force_directed is not None:
            self.set_directed(force_directed)
        else:
            # Try to guess directed or not
            y = self.adj()
            adj_symmetric = (y!=y.T).nnz == 0
            adj_symmetric |= sp.sparse.triu(y,1).nnz == 0
            adj_symmetric |= sp.sparse.tril(y,-1).nnz == 0
            if adj_symmetric:
                self.log.info('Guessing undirected graph for: %s' % expe.corpus)
                self.set_directed(False)

            if y.max() > 1:
                self.log.critical('Multiple edges in the graph..')

        if not 'weights' in data.ep:
            if 'weight' in data.ep:
                self.log.critical('Already Weight propertye in the graph, need to check ')
                raise NotImplementedError
            elif 'value' in data.ep:
                weights = data.ep['value'].copy()
                #w.a = (2**0.5)**w.a # exponentiate
                #weights.a = np.round((weights.a+1)**2) # squared
                weights.a = np.round(weights.a *10) # assume ten authors by paper for (collaboration networks)
                data.ep['weights'] = weights.copy('int')
            else:
                weights = data.new_ep("int")
                weights.a = 1
                data.ep['weights'] = weights
        else:
            weights = data.ep.weights

        if self.expe.get('shift_w'):
            shift = self.expe.get('shift_w')
            if str(shift).startswith('linear'):
                _, shift = shift.split('_')
                #shift = weights.a.max()
                shift = int(shift)
                expe['shift_w'] = shift
                weights.a = shift*weights.a +shift
            else:
                shift = int(self.expe['shift_w'])
                weights.a += shift



    @classmethod
    def _extract_data_file(cls, expe, corpus=None):
        corpus = corpus or {}
        input_path = expe._input_path

        if not os.path.exists(input_path):
            cls.log.error("Corpus `%s' Not found." % (input_path))
            print('please run "fetch_networks"')
            cls.data = None
            return

        if expe.corpus.endswith('.gt'):
            corpus_name = expe.corpus[:-len('.gt')]
        else:
            corpus_name = expe.corpus

        # DB integration ?
        if corpus_name.startswith(('generator', 'graph')):
            ext = 'graph'
            basename = os.path.join(input_path, 't0.' + ext)
            fun = 'parse_dancer'
        elif corpus_name.startswith(('bench1')):
            raise NotImplementedError()
        elif corpus_name.startswith('facebook'):
            basename = 'edges'
            fn = os.path.join(input_path, '0.' + ext)
            fun = 'parse_edges'
        elif corpus_name.startswith(('manufacturing',)):
            ext = 'csv'
            basename = os.path.join(input_path, corpus_name +'.'+ ext)
            fun = 'parse_csv'
        elif corpus_name.startswith(('fb_uc', 'emaileu')):
            ext = 'txt'
            basename = os.path.join(input_path, corpus_name +'.'+ ext)
            fun = 'parse_tnet'
        elif corpus_name.startswith(('blogs','propro', 'euroroad')):
            ext = 'dat'
            basename = os.path.join(input_path, corpus_name +'.'+ ext)
            fun = 'parse_dat'
        else:
            raise ValueError('Which corpus to Load; %s ?' % corpus_name)

        fn = os.path.join(input_path, basename)
        if os.path.isfile(fn) and os.stat(fn).st_size == 0:
            cls.log.warning('Doh, Corpus file is empty at: %s' % fn)
            return

        parse_fun = getattr(cls, fun)

        N = corpus.get('nodes')
        E = corpus.get('edges')

        g = gt.Graph(directed=True)
        #g.add_vertex(N)

        weights = g.new_edge_property("int") # g.new_ep
        labels = g.new_vertex_property("string") # g.new_vp
        clusters = g.new_vertex_property("int") # g.new_vp

        ### Slower but weight correct
        for obj in parse_fun(fn):

            n_nodes = g.num_vertices()

            if isinstance(obj, tuple):
                i, j, w, f = obj
                if max(i,j) >= n_nodes:
                    g.add_vertex(max(i,j) - n_nodes + 1)
                edge = g.edge(i,j)
                if not edge:
                    # keep a binary networks in the graph structure.
                    # Number of edge are in the {weights} property.
                    edge = g.add_edge(i, j)

                weights[edge] = w + weights[edge]
            else:
                v = obj.pop('index')
                if v >= n_nodes:
                    g.add_vertex(v - n_nodes + 1)

                if 'label' in obj:
                    labels[v] = obj['label']
                if 'cluster' in obj:
                    clusters[v] = obj['cluster']

        ### Faster but weigth wrong
        #_edges = []
        #_index = []
        #_label = []
        #for obj in parse_fun(fn):
        #    if isinstance(obj, tuple):
        #        i, j, w, f = obj
        #        _edges.append([i,j,w])
        #    else:
        #        _index.append(obj['index'])
        #        _label.append(obj['label'])

        #g.add_edge_list(_edges, eprops=[weights])
        #if _label:
        #    labels.a[_index] = _label

        g.edge_properties['weights'] = weights # g.ep
        g.vertex_properties['labels'] = labels # g.vp
        g.vertex_properties['clusters'] = clusters

        # If nolabels remove that property
        n_label = 0
        for v in range(g.num_vertices()):
            if g.vp['labels']:
                n_label += 1
                break

        if n_label == 0:
            del g.vp['labels']
        # If all clusters are zeros, consider no label information
        if g.vp['clusters'].a.sum() == 0:
            del g.vp['clusters']

        # Remove first index in case of indexation start at 1
        zero_degree = g.vertex(0).out_degree() + g.vertex(0).in_degree()
        if zero_degree == 0:
            cls.log.debug('shifting the graph. (-1 to vertex index)')
            g.remove_vertex(0)

        _N = g.num_vertices()
        _E = g.num_edges()

        if N and N != _N:
            cls.log.warning('Number of nodes differs, doc: %d, code: %d' % (N, _N))
        if E and E != _E:
            cls.log.warning('Number of edges differs, doc: %d, code: %d' % (E, _E))

        return g


    @classmethod
    def _clean_data_konect(cls, expe, data):
        cls.log.info('Cleaning konect graph %s...' % expe.corpus)

        g = data
        edges = g.get_edges()
        y = gt.spectral.adjacency(g)
        parallel = gt.stats.label_parallel_edges(g, mark_only=True)

        weights = g.new_ep('int')

        if 'weight' in g.ep:

            w = g.ep.weight

            if w.a.sum() == len(w.a):
                # Assume parrallel edge
                # Building the real weights
                weights.a = 1
                for _ix in np.where(parallel.a == 0)[0]:
                    ix = np.where(edges[:,2]==_ix)[0][0]
                    i, j = edges[ix, :2]
                    weights[i,j] += len(g.edge(i,j, all_edges=True))


                del g.ep['weight']
            elif w.a.max() == 1 and w.a.min() == -1:
                for i,j, ix in edges:
                    weights[i,j] += w[i,j]

                if weights.a.min() < 0:
                    # Remove negative edges
                    cls.log.warning('Negative weights here? removing edges...')
                    print('Number of negative edges: %d' % (weights.a < 0).sum() )
                    #print('Number of negative edges: %d' % (weights.a < 0).sum() )
                    #edge_f = g.new_ep('bool')
                    #edge_f.a = 1
                    #edge_f.a[weights.a < 0] = 0
                    #g.set_edge_filter(edge_f)
                    #g.purge_edges()
                    #g.clear_filters()
                    weights.a[weights.a < 0] = 1

                if (weights.a==0).sum() > 0:
                    # Remove empty edges
                    cls.log.warning('Removing zeros weighted edges...')
                    print('Number of zeros edges: %d' % (weights.a == 0).sum() )
                    #edge_f = g.new_ep('bool')
                    #edge_f.a = 1
                    #edge_f.a[weights.a == 0] = 0
                    #g.set_edge_filter(edge_f)
                    #g.purge_edges()
                    #g.clear_filters()
                    weights.a[weights.a == 0] = 1

            elif y.max() == 1:
                # Assume weight are the real one, copy
                weights = g.ep.weight.copy('int')
                cls.log.warning('Weight regression to int values.')

            elif y.max() > 1:
                raise NotImplementedError('Multi edge + weight property for %s, manual check needed' % expe.corpus)
            else:
                raise NotImplementedError('Edge propertie unknown for %s, manual check needed' % expe.corpus)

        elif 'weights' in g.ep:
            self.log.critical('Already weights property for %s, manual check needed' % expe.corpus)
            raise NotImplementedError
        elif y.max() > 1:
            # If multiple edge detected, set the weights.
            weights.a = 1
            for _ix in np.where(parallel.a == 0)[0]:
                ix = np.where(edges[:,2]==_ix)[0][0]
                i, j = edges[ix][:2]
                weights[i,j] += len(g.edge(i,j, all_edges=True))
        else:
            # Assume unweighted network
            weights.a = 1


        # Remove parallel edges
        #gt.stats.remove_parallel_edges(g) # dont't resize .a !
        g.set_edge_filter(parallel, inverted=True)
        g.purge_edges()
        g.clear_filters()

        g.shrink_to_fit()
        g.ep['weights'] = weights
        return g


    @classmethod
    def _resolve_filename(cls, expe):
        input_path = expe._input_path

        if expe.corpus.endswith('.gt'):
            basename = expe.corpus
        else:
            basename = expe.corpus + '.gt'

        fn = os.path.join(input_path, basename)
        return fn

    @classmethod
    def _save_data(cls, fn, data):
        driver = data.save
        return super()._save_data(fn, None, driver=driver)

    @classmethod
    def _load_data(cls, fn):
        driver = gt.load_graph
        return super()._load_data(fn, driver=driver)


    @classmethod
    def from_expe(cls, expe, corpus=None, load=True, save=True):
        if '_force_load_data' in expe:
            load = expe._force_load_data
        if '_force_save_data' in expe:
            save = expe._force_save_data

        input_path = cls.get_input_path(expe)

        data = None
        fn = cls._resolve_filename(expe)
        target_file_exists = os.path.exists(fn)

        if load is False or not target_file_exists:

            # Data loading Strategy
            if not data:
                try:
                    # Load from graph-tool Konnect repo
                    from graph_tool import collection
                    data = gt.load_graph(collection.get_data_path(expe.corpus))
                    os.makedirs(os.path.join(input_path), exist_ok=True)
                except FileNotFoundError as e:
                    pass
                except Exception as e:
                    cls.log.error("Error in loading corpus `%s': %s" % (expe.corpus, e))
                    raise e

            if not data:
                try:
                    from urllib.error import HTTPError
                    # Load from graph-tool Konnect site
                    data = gt.collection.konect_data[expe.corpus]
                    data = cls._clean_data_konect(expe, data)
                    os.makedirs(os.path.join(input_path), exist_ok=True)
                except HTTPError:
                    pass
                except Exception as e:
                    cls.log.error("Error in loading corpus `%s': %s" % (expe.corpus, e))
                    raise e

            if not data:
                # Load manually from file
                data = cls._extract_data_file(expe, corpus=corpus)

            if save:
                # ===== save ====
                cls._save_data(fn, data)
        else:
            # ===== load ====
            data = cls._load_data(fn)

        return cls(expe, data, corpus=corpus)


    #
    # Get Properties
    #

    def edge(self, i,j):
        return self.data.edge(i,j)

    def weight(self, i,j):
        w = self.data.ep['weights']
        if self.edge(i,j):
            return w[i,j]
        else:
            return 0

    def label(self, v):
        l = self.data.vp['labels']
        return l[v]

    def cluster(self, v):
        c = self.data.vp['clusters']
        return c[v]

    def num_neighbors(self):
        neigs = []
        for v in range(self.num_nodes()):
            _v = self.data.vertex(v)
            neigs.append(len(list(_v.all_neighbors())))
        return np.asarray(neigs)

    def is_directed(self):
        return self.data.is_directed()
    def is_symmetric(self):
        return not self.data.is_directed()

    def getN(self):
        return self.num_nodes()

    def get_validset_ratio(self):

        if 'validset_ratio' in self.expe:
            validset_ratio = self.expe['validset_ratio']
            validset_ratio = float(validset_ratio) / 100
        else:
            validset_ratio = 0

        return validset_ratio

    def get_training_ratio(self):

        if 'training_ratio' in self.expe:
            training_ratio = self.expe['training_ratio']
            training_ratio = float(training_ratio) / 100
        else:
            training_ratio = 0

        return training_ratio

    def get_testset_ratio(self):

        if 'testset_ratio' in self.expe:
            testset_ratio = self.expe['testset_ratio']
            testset_ratio = float(testset_ratio) / 100
        else:
            testset_ratio = 0

        validset_ratio = self.get_validset_ratio()

        # Validation ratio is a ratio took on the remaining data (eta_ratio...)
        testset_ratio = testset_ratio * (1 + validset_ratio)

        return testset_ratio

    def num_nodes(self):
        return self.data.num_vertices()

    def num_edges(self):
        return self.data.num_edges()

    def num_nnz(self):
        N = self.data.num_vertices()
        sym_pt = 1 if self.is_directed() else 2
        if hasattr(self, 'data_test'):
            T = N*(N-1)/sym_pt - int(self.data_test.sum()/sym_pt)
        else:
            T = N*(N-1)/sym_pt

        return T

    def num_nnzsum(self):
        return self.data.ep['weights'].a.sum()

    def num_mnb(self):
        ''' Minibatche size for the sampling. '''
        return int(self.num_nodes() * self._zeros_set_len * float(self.expe['sampling_coverage']))

    def diameter(self):
        diameter, end_points = gt.topology.pseudo_diameter(self.data)
        return int(diameter)

    def density(self):
        e = self.data.num_edges()
        v = self.data.num_vertices()
        t = v * (v-1)
        if not self.data.is_directed():
            t /= 2
        return e / t

    def modularity(self):
        if not 'labels' in self.data.vp:
            return None
        labels = self.data.vp['labels']

        if 'weights' in self.data.ep:
            weights = self.data.ep['weights']
        else:
            weights = None

        mod = gt.inference.modularity(self.data, labels, weights)
        return mod

    def clustering_coefficient(self):
        clust_coef, std = gt.clustering.global_clustering(self.data)
        return clust_coef

    def net_type(self):
        weights = self.data.ep['weights'].a
        return 'm/std/mn/mx: %.1f %.1f %d %d' % (weights.mean(),
                                                      weights.std(),
                                                      weights.min(),
                                                      weights.max())

    def feat_len(self):
        weights = self.data.ep['weights'].a
        return len(np.unique(weights))

    @staticmethod
    def _prop(g):
        n, e, y = g.num_vertices(), g.num_edges(), gt.spectral.adjacency(g)
        sum = y.sum()
        #sl = np.diag(y.A).sum()
        sl = y[(np.arange(n),np.arange(n))].sum()
        print(g)
        print('N: %d, E: %d, adj.sum: %d, adj.selfloop: %d' % (n,e,sum,sl))
        print('edges shape', g.get_edges().shape)
        print('Vertex prop', g.vp)
        print('Edge prop', g.ep)


    #
    # Transform
    #

    def reverse_filter(self):
        filter, is_inv = self.data.get_edge_filter()
        self.data.set_edge_filter(filter, inverted=not is_inv)

    def adj(self):
        ''' Returns a sparse Adjacency matrix.

            Notes
            -----
            index of the adjacency are reversed from np convention i:line, j:column
        '''
        return gt.spectral.adjacency(self.data).T

    def laplacian(self):
        ''' Returns a sparse Laplacian matrix. '''
        return gt.spectral.laplacian(self.data)


    def binarize(self):
        # Or set a propertie with gt.stats.label_parallel_edges(g)
        return gt.stats.remove_parallel_edges(self.data)

    def set_directed(self, directed=True):
        # Modifying the adjency matrix don't modify the gt graph
        #y = self.adj()
        #(y!=y.T).nnz == 0
        self.data.set_directed(directed)

    def remove_self_loops(self):
        # Or set a propertie with gt.stats.remove_self_loops(g)
        return gt.stats.remove_self_loops(self.data)

    def sample(self, N):
        ''' Reduce randomly  the number of nodes of the graph. '''
        n_to_remove = self.data.num_vertices() - N
        ids = np.random.randint(0, self.data.num_vertices(), n_to_remove)
        self.data.remove_vertex(ids)
        self.data.shrink_to_fit()



    #
    # Test/Validation set
    #

    def make_testset(self, diag_off=1, set_filter=True):
        ''' make the testset as a edge propertie of the graph.
            If {set_filter} is True, activate the filtering property on the graph

            The algo try to build a equillibrite testset between edge and non-edge with
            some uncertanty for undirected graph.

            Notes
            -----
            For Undirected graph, the sparse matrice is symmetrized, so the value of
            edges can be cummulated from the two inverted sample. Thus the ratio for the
            non-edges is reduce to 1/3 of the edge ratio to be fair.

            This operation can be sub-optimal since it compare array of size N**2.

            Warning: if set_filter=false, wiehgt is not set in the testset matrix.
        '''

        testset_ratio = self.get_testset_ratio()
        if not testset_ratio:
            raise ValueError('Testset ratio not understood : %s' % testset_ratio)

        g = self.data
        y = self.adj()
        N = g.num_vertices()
        E = g.num_edges()
        is_directed = not self.is_symmetric()
        symmetric_scale = 1 if is_directed else 2
        size = N**2

        #validset = lil_matrix((N,N))
        testset = lil_matrix((N,N), dtype=bool)
        n = int(E * testset_ratio) # number of edge
        nz = int(n *float(self.expe.get('zeros_ratio', 1)))  # number of non-links

        edges = g.get_edges().astype(int)
        i, j = edges[:, 0:2].T

        # Sampling edges
        ix = np.random.choice(E, n, replace=False)
        ix = np.array((i[ix], j[ix]))
        #testset[ix[0], ix[1]] = 1

        # Sampling non-edges
        #
        # Ignore the diagonal => include the diag indices
        # in (i,j) like if it they was selected in the testset.
        # If Undirected also force ignore the lower part of the adjacency matrix.
        if is_directed:
            i = np.hstack((i, np.arange(N)))
            j = np.hstack((j, np.arange(N)))
        else:
            _i = i
            i = np.hstack((i, j, np.arange(N)))
            j = np.hstack((j, _i, np.arange(N)))

        edge_set = set(np.ravel_multi_index((i,j), y.shape))

        nonlinks_set = set()
        max_loop = 150
        cpt = 0
        # Add the probability to sample symmetric edge...
        zeros_len = int(nz + int(not is_directed)*nz/size)
        while len(nonlinks_set) < zeros_len or cpt >= max_loop:
            nnz = zeros_len - len(nonlinks_set)
            nonlinks_set.update(set(np.random.choice(size, nnz)))
            nonlinks_set -= edge_set
            cpt+=1

        self.log.debug('Number of iteration for sampling non-edges: %d' % cpt)
        if cpt >= max_loop:
            self.log.warning('Sampling tesset nonlinks failed:')
            print('desires size: %d, obtained: %s' % (zeros_len, len(nonlinks_set)))

        jx = np.array(np.unravel_index(list(nonlinks_set), (N,N)))

        #k = np.ravel_multi_index((i,j), y.shape)
        #jx = np.random.choice(size, int(nz/symmetric_scale), replace=False)

        #kind = np.bincount(k, minlength=size).astype(bool)
        #jind = np.bincount(jx, minlength=size).astype(bool)

        ## xor on bool faster ?!
        #_n = (jind & (kind==0)).sum()
        #max_loop = 100
        #cpt = 0
        #while _n < int(nz/symmetric_scale) or cpt > max_loop:
        #    jx = np.random.choice(size, int(nz/symmetric_scale) - _n, replace=False)
        #    jind = jind|np.bincount(jx, minlength=size).astype(bool)
        #    _n = (jind & (kind==0)).sum()
        #    cpt += 1
        #    #print(_n, n)
        #self.log.debug('Number of iteration for sampling non-edges: %d'%cpt)

        #jx = np.arange(size)[jind & (kind==0)]
        #jx = np.array(np.unravel_index(jx, (N,N)))


        # Actually settings the testset entries
        testset[ix[0], ix[1]] = 1
        testset[jx[0], jx[1]] = 1
        if not is_directed:
            testset[ix[1], ix[0]] = 1
            testset[jx[1], jx[0]] = 1


        data_test_w = np.transpose(testset.nonzero())
        #self.reverse_filter()
        weights = []
        for i,j in data_test_w:
            weights.append(self.weight(i,j))
        #self.reverse_filter()
        data_test_w = np.hstack((data_test_w, np.array(weights)[None].T))

        if set_filter:
            # Also set the weight here...
            testset_filter = g.new_ep('bool')
            testset_filter.a = 1
            self.log.info('Looping to set the testset edge eproperties (graph-tool)...')
            for i, j in ix.T:
                testset_filter[i,j] = 0
            g.set_edge_filter(testset_filter)

        self._links_testset = ix
        self._nonlinks_testset = jx
        self.data_test = testset
        self.data_test_w = data_test_w

        # if we learn with a training_ratio < 1
        # we just ignore some random point.
        training_ratio = self.get_training_ratio()
        if training_ratio > 0 and training_ratio < 1:
            # remove edges
            self.data.set_fast_edge_removal()

            edges = self.data.get_edges()
            weights = self.data.ep['weights']
            n_extra = len(edges) - training_ratio*len(edges)

            edges_to_remove = np.random.choice(len(edges), int(n_extra), replace=False)
            for pos in edges_to_remove:
                i,j = edges[pos, :2]
                e = g.edge(i,j)
                g.remove_edge(e)

            self.data.set_fast_edge_removal(fast=False)
        return

    def _check(self):
        ''' Warning: will clear the filter! '''
        expe = self.expe
        g = self.data
        self._prop(g)

        y = self.adj()
        N = g.num_vertices()
        E = g.num_edges()

        # check no self-loop
        assert(y.diagonal().sum() == 0)
        # check directed/undirected is consistent
        is_directed = not self.is_symmetric()
        adj_symmetric = (y!=y.T).nnz == 0
        assert(adj_symmetric != is_directed)
        # check is not bipartite
        assert(not gt.topology.is_bipartite(g))

        if 'testset_ratio' in expe:
            testset_ratio = self.get_testset_ratio()
            filter = g.get_edge_filter()
            g.clear_filters()
            y = self.adj()
            n = int(g.num_edges() * testset_ratio)
            nz = int(n *float(self.expe.get('zeros_ratio', 1)))
            ix = self._links_testset
            jx = self._nonlinks_testset
            # check links testset
            assert(np.all(y[ix[0], ix[1]] == 1))
            # check non-links testset
            assert(np.all(y[jx[0], jx[1]] == 0))

            print('Size of testset: expected: %d, obtained: %d' % (n+nz, self.data_test.sum()))
            print('number of links: %d' % (len(ix[0])))
            print('number of non-links: %d' % (len(jx[0])))
            if filter[0]:
                g.set_edge_filter(filter[0], inverted=filter[1])

        return

    def make_noise(self):
        expe = self.expe
        g = self.data
        is_directed = g.is_directed()

        symmetric_scale = 1 if is_directed else 2

        y = self.adj()
        E = g.num_edges()
        N = g.num_vertices()
        T = N*(N-1)/symmetric_scale

        weights = g.ep['weights']

        ratio = 1/float(expe['noise'])
        #ts = self.get_testset_ratio()
        #if ts:
        #    ratio -= 2*ts*E/T
        #    if ratio < 0:
        #        raise ValueError('Negative prior for noise ration')
        g.set_fast_edge_removal()

        edges = g.get_edges()
        edges_to_remove = np.random.choice(E, int(ratio*E), replace=False)
        _removed = set()
        for pos in edges_to_remove:
            i,j = edges[pos, :2]
            w = weights[i,j]
            if w > 1:
                weights[i,j] = w-1
            else:
                e = g.edge(i,j)
                _removed.add(e)
                g.remove_edge(e)

        num_to_add = ratio*E*2
        cpt = 0
        while cpt < num_to_add:
            i, j = np.random.randint(0,N, 2)
            if g.edge(i,j):
                continue
            else:
                e = g.add_edge(i,j)
                if e in _removed:
                    continue
                weights[e] = 1
                cpt+=1

        return g.set_fast_edge_removal(fast=False)

    def __iter__(self):

        # Get the sampling strategy:
        chunk = self.expe.get('chunk', 'stratify')

        if chunk == 'stratify':
            return self._sample_stratify()
        elif chunk == 'sparse':
            return self._sample_sparse()
        else:
            raise NotImplementedError('Unkonw sampling strategy: %s' % chunk)


    def _sample_stratify(self):
        ''' Sample with node replacement.
            But edges in a minibatch is unique.

            Returns
            ------

            yield ether: a triplet (source, target, weight), that can be of two kinf
            1. at a new minibatch (mnb):
                * source: str -- a str that identify from which subset the mnb comes.
                * target: int -- the target node of the mnb.
                * weight: int -- the probability of to sample a element in the current set/mnb
                          in the corpus/dataset.
            2. an obeservation (edges):
                * source: int -- source node.
                * target: int -- target node.
                * weight: int -- edge weight.
        '''
        expe = self.expe
        g = self.data
        is_directed = g.is_directed()

        symmetric_scale = 1 if is_directed else 2 # number of set containg each edges

        y = self.adj()
        E = g.num_edges()
        N = g.num_vertices()
        T = N*(N-1)/symmetric_scale

        weights = g.ep['weights']

        ### Chunk prior probability
        set_len = 2 + is_directed
        # uniform between links and non-links
        zero_prior = float(expe.get('zeros_set_prob', 0.5))
        set_prior = [zero_prior] + [(1-zero_prior)/(set_len-1)]*(set_len-1)

        ### Sampling prior strategy
        zeros_set_len = expe.get('zeros_set_len', 10)
        self._zeros_set_len = int(zeros_set_len)
        mask = np.ones((N,2), dtype=int) * self._zeros_set_len

        ### Build the array of neighbourhood for each nodes
        self.log.debug('Building the neighbourhood array...')
        neigs = []
        mask_pos = []
        for v in range(N):
            _out = np.array([int(_v) for _v in g.vertex(v).out_neighbors()])
            _in = np.array([int(_v) for _v in g.vertex(v).in_neighbors()])
            neigs.append([_out, _in])
            mask_pos.append( [list(range(self._zeros_set_len))]*2 )


        for _mnb in range(self.num_mnb()):

            # Pick a node
            node = np.random.randint(0,N)

            # Pick a set and yield a minibatch
            set_index = np.random.choice(set_len, 1, p=set_prior)

            if set_index == 0:
                out_e = neigs[node][0]
                if N-len(out_e) > 0 and len(out_e) > 0:

                    node_info = {'vertex':node, 'direction':0}
                    yield str(set_index), node_info, 1/symmetric_scale*set_len*mask[node, 0]
                    # Sample from non-links

                    zero_samples = np.arange(N)

                    zero_samples[out_e] = -1 # don't sample in edge
                    zero_samples[node] = -1 # don't sample in self loops
                    #zero_samples[self.data_test[node, :].nonzero()[1]] = -1 # don't sample in testset
                    zero_samples = zero_samples[zero_samples >0]

                    zero_samples = np.random.choice(zero_samples,
                                                     int(np.ceil(len(zero_samples)/mask[node, 0])),
                                                     replace=False)
                    # Weaker results !
                    #step = int(np.ceil(len(zero_samples)/mask[node, 0]))
                    #if len(mask_pos[node][0]) == 0:
                    #    continue
                    #    zero_samples = np.random.choice(zero_samples,
                    #                                    int(np.ceil(len(zero_samples)/mask[node, 0])),
                    #                                    replace=False)
                    #else:
                    #    _p = np.random.choice(mask_pos[node][0],1).item()
                    #    mask_pos[node][0].remove(_p)
                    #    zero_samples = zero_samples[_p:_p+step]

                    if len(zero_samples) == 0:
                        continue

                    for target in zero_samples:
                        yield node, target, 0

                in_e = neigs[node][1]
                if N-len(in_e) > 0 and len(in_e) > 0:
                    node_info = {'vertex':node, 'direction':1}
                    yield str(set_index), node_info, 1/symmetric_scale*set_len*mask[node, 1]

                    zero_samples = np.arange(N)
                    if len(in_e) > 0:
                        zero_samples[in_e] = -1
                    zero_samples[node] = -1
                    #zero_samples[self.data_test[:, node].nonzero()[0]] = -1
                    zero_samples = zero_samples[zero_samples >0]

                    zero_samples = np.random.choice(zero_samples,
                                                    int(np.ceil(len(zero_samples)/mask[node, 1])),
                                                    replace=False)
                    # Weaker results !
                    #step = int(np.ceil(len(zero_samples)/mask[node, 1]))
                    #if len(mask_pos[node][1]) == 0:
                    #    continue
                    #    zero_samples = np.random.choice(zero_samples,
                    #                                    int(np.ceil(len(zero_samples)/mask[node, 1])),
                    #                                    replace=False)
                    #else:
                    #    _p = np.random.choice(mask_pos[node][1],1).item()
                    #    mask_pos[node][1].remove(_p)
                    #    zero_samples = zero_samples[_p:_p+step]

                    if len(zero_samples) == 0:
                        continue

                    for target in zero_samples:
                        yield target, node, 0

            elif set_index == 1:
                # Sample from links
                if len(neigs[node][0]) == 0:
                    continue

                node_info = {'vertex':node, 'direction':0}
                yield str(set_index), node_info, 1/symmetric_scale*N*set_len

                for target in neigs[node][0]:
                    yield node, target, weights[node, target]
            elif set_index == 2:
                # Sample from links
                if len(neigs[node][1]) == 0:
                    continue

                node_info = {'vertex':node, 'direction':1}
                yield str(set_index), node_info, 1/symmetric_scale*N*set_len

                for target in neigs[node][1]:
                    yield target, node, weights[target, node]
            else:
                raise ValueError('Set index error: %s' % set_index)


    def _sample_sparse(self):
        ''' Sample with node replacement.
        '''
        expe = self.expe
        g = self.data
        is_directed = g.is_directed()

        symmetric_scale = 1 if is_directed else 2 # number of set containg each edges

        y = self.adj()
        E = g.num_edges()
        N = g.num_vertices()
        T = N*(N-1)/symmetric_scale

        weights = g.ep['weights']

        ### Chunk prior probability
        set_len = 2 + is_directed
        # uniform between links and non-links
        zero_prior = float(expe.get('zeros_set_prob', 0.5))
        set_prior = [zero_prior] + [(1-zero_prior)/(set_len-1)]*(set_len-1)

        ### Sampling prior strategy
        zeros_set_len = expe.get('zeros_set_len', 10)
        self._zeros_set_len = int(zeros_set_len)
        mask = np.ones((N,2), dtype=int) * self._zeros_set_len

        ### Build the array of neighbourhood for each nodes
        self.log.debug('Building the neighbourhood array...')
        neigs = []
        mask_pos = []
        for v in range(N):
            _out = np.array([int(_v) for _v in g.vertex(v).out_neighbors()])
            _in = np.array([int(_v) for _v in g.vertex(v).in_neighbors()])
            neigs.append([_out, _in])
            mask_pos.append( [list(range(self._zeros_set_len))]*2 )


        for _mnb in range(self.num_mnb()):

            # Pick a node
            node = np.random.randint(0,N)

            # Pick a set and yield a minibatch
            set_index = np.random.choice(set_len, 1, p=set_prior)

            if set_index == 0:
                out_e = neigs[node][0]
                if N-len(out_e) > 0 and len(out_e) > 0:

                    node_info = {'vertex':node, 'direction':0}
                    yield str(set_index), node_info, (N-len(out_e)-1)* mask[node,0]* symmetric_scale
                    # Sample from non-links

                    zero_samples = np.arange(N)

                    zero_samples[out_e] = -1 # don't sample in edge
                    zero_samples[node] = -1 # don't sample in self loops
                    #zero_samples[self.data_test[node, :].nonzero()[1]] = -1 # don't sample in testset
                    zero_samples = zero_samples[zero_samples >0]
                    zero_samples = np.random.choice(zero_samples,
                                                     int(np.ceil(len(zero_samples)/mask[node, 0])),
                                                     replace=False)
                    # Weaker results !
                    #step = int(np.ceil(len(zero_samples)/mask[node, 0]))
                    #if len(mask_pos[node][0]) == 0:
                    #    zero_samples = np.random.choice(zero_samples,
                    #                                    int(np.ceil(len(zero_samples)/mask[node, 0])),
                    #                                    replace=False)
                    #else:
                    #    _p = np.random.choice(mask_pos[node][0],1).item()
                    #    mask_pos[node][0].remove(_p)
                    #    zero_samples = zero_samples[_p:_p+step]

                    if len(zero_samples) == 0:
                        continue

                    for target in zero_samples:
                        yield node, target, 0

                in_e = neigs[node][1]
                if N-len(in_e) > 0 and len(in_e) > 0:
                    node_info = {'vertex':node, 'direction':1}
                    yield str(set_index), node_info, (N-len(in_e)-1)* mask[node,1]* symmetric_scale

                    zero_samples = np.arange(N)
                    if len(in_e) > 0:
                        zero_samples[in_e] = -1
                    zero_samples[node] = -1
                    #zero_samples[self.data_test[:, node].nonzero()[0]] = -1
                    zero_samples = zero_samples[zero_samples >0]
                    zero_samples = np.random.choice(zero_samples,
                                                    #int(np.ceil(len(zero_samples)/(10*len(in_e)))),
                                                    int(np.ceil(len(zero_samples)/mask[node, 1])),
                                                    replace=False)
                    # Weaker results !
                    #step = int(np.ceil(len(zero_samples)/mask[node, 1]))
                    #if len(mask_pos[node][1]) == 0:
                    #    zero_samples = np.random.choice(zero_samples,
                    #                                    int(np.ceil(len(zero_samples)/mask[node, 1])),
                    #                                    replace=False)
                    #else:
                    #    _p = np.random.choice(mask_pos[node][1],1).item()
                    #    mask_pos[node][1].remove(_p)
                    #    zero_samples = zero_samples[_p:_p+step]

                    if len(zero_samples) == 0:
                        continue

                    for target in zero_samples:
                        yield target, node, 0

            elif set_index == 1:
                # Sample from links
                if len(neigs[node][0]) == 0:
                    continue

                node_info = {'vertex':node, 'direction':0}
                yield str(set_index), node_info, symmetric_scale*N*len(neigs[node][0])

                for target in neigs[node][0]:
                    yield node, target, weights[node, target]
            elif set_index == 2:
                # Sample from links
                if len(neigs[node][1]) == 0:
                    continue

                node_info = {'vertex':node, 'direction':1}
                yield str(set_index), node_info, symmetric_scale*N*len(neigs[node][1])

                for target in neigs[node][1]:
                    yield target, node, weights[target, node]
            else:
                raise ValueError('Set index error: %s' % set_index)







#!/usr/bin/python3
#import pyximport; pyximport.install(pyimport = True)
import sys, os
from pymake import GramExp


''' A Command line controler of Pymake '''

def bootstrap():

    env = dict(os.environ)
    pwd = env.get('PWD')
    ## change directory if asked
    if '-cd' in sys.argv:
        i = sys.argv.index('-cd')
        p = sys.argv[i+1]
        os.chdir(p)

        # Debug ?
        sys.argv.pop(i+1)
        sys.argv.pop(i)
        pwd = os.path.join(pwd, p)
        env['PWD'] = pwd
    else:
        # @debug this, I dont know whyiam in lib/package sometimes, annoying !
        os.chdir(env.get('PWD'))


    ## Search in the project and current repo. Awesome !
    sys.path.insert(0, pwd + '/.')
    sys.path.insert(0, pwd + '/..')

    return env



def main():

    env = bootstrap()
    GramExp.setenv(env)

    zymake = GramExp.zymake()
    zyvar = zymake._conf

    if zyvar.get('simulate') and ( not zyvar['_do'] in ['run', 'runpara', 'path'] or not zyvar.get('script')):
        # same as show !
        zymake.simulate()


    lines = None
    line_prefix = ''
    if zyvar['_do'] == 'init':
        zymake.init_folders()
        exit()
    else:
        if (zyvar['_do'] or zyvar.get('do_list')) and not GramExp.is_pymake_dir():
            print('fatal: Not a pymake directory: %s not found.' % (GramExp._cfg_name))
            try:
                os.remove('.pmk-db.db')
            except FileNotFoundError:
                pass
            exit(10)

    if zyvar['_do'] == 'update':
        zymake.update_index()
    elif zyvar['_do'] == 'show':
        zymake.simulate()
    elif zyvar['_do'] ==  'run':
        lines = zymake.execute()
    elif zyvar['_do'] == 'runpara':
        is_distrib = zyvar.get('_net')
        if is_distrib:
            if is_distrib != True:
                nhosts = int(is_distrib)
            else:
                nhosts = None
            lines = zymake.execute_parallel_net(nhosts)
        else:
            lines = zymake.execute_parallel()
        zymake.pushcmd2hist()
    elif zyvar['_do'] == 'cmd':
        lines = zymake.make_commandline()
    elif zyvar['_do'] == 'path':
        lines = zymake.make_path(ext=zyvar.get('_ext'), status=zyvar.get('_status'))
    elif zyvar['_do'] == 'hist':
        lines = zymake.show_history()
    elif zyvar['_do'] == 'diff':
        lines = zymake.show_diff()
    elif zyvar['_do'] == 'notebook': # @Todo
        lines = zymake.notebook()
    else: # list things

        if not 'do_list' in zyvar and zyvar['_do']:
            raise ValueError('Unknown command : %s' % zyvar['_do'])

        if 'spec' == zyvar.get('do_list'):
            print (zymake.spectable())
        elif 'model' == zyvar.get('do_list'):
            print (zymake.modeltable())
        elif 'script' == zyvar.get('do_list'):
            print(zymake.scripttable())
        elif 'model_topos' == zyvar.get('do_list'):
            print (zymake.modeltable(_type='topos'))
        elif 'spec_topo' ==  zyvar.get('do_list'):
            print (zymake.spectable_topo())
        elif 'topo' ==  zyvar.get('do_list'):
            print (zymake.alltable_topo())
        else:
            print(zymake.help_short())
            if zyvar.get('do_list'):
                print ('Unknow options %s ?' % zyvar.get('do_list'))
        exit()

    if lines is None:
        # catch signal ?
        exit()

    if 'script' in zyvar:
        lines = [' '.join((line_prefix, l)).strip() for l in lines]

    zymake.simulate(halt=False, file=sys.stderr)
    print('\n'.join(lines), file=sys.stdout)

if __name__ == '__main__':
    main()




import numpy as np

from pymake import ExpeFormat
from pymake.manager import ModelManager, FrontendManager

from pymake.util.utils import colored



''' Build and Stats for Networds data '''



class Net(ExpeFormat):

    _default_expe = dict(
        _spec = 'data_net_all',
        _data_type = 'networks'
    )

    def __call__(self):
        return self.stats()

    def stats(self, fmt='simple'):
        ''' Show data stats '''
        expe = self.expe
        frontend = FrontendManager.load(expe)

        if self.is_first_expe():
            # Warning order sensitive @deprecated Table.
            #corpuses = self.specname(self.gramexp.get_set('corpus'))
            corpuses = self.specname(self.gramexp.get_list('corpus'))
            Meas = ['num_nodes', 'num_edges', 'density',
                    'is_directed', 'modularity', 'diameter', 'clustering_coefficient', 'net_type', 'feat_len']
            Meas_ = ['nodes', 'edges', 'density',
                     'directed', 'modularity', 'diameter', 'cluster-coef', 'weights', 'feat-len']
            Table = np.zeros((len(corpuses), len(Meas))) * np.nan
            Table = np.column_stack((corpuses, Table))
            self.D.Table = Table
            self.D.Meas = Meas
            self.D.Meas_ = Meas_

        Table = self.D.Table
        Meas = self.D.Meas

        for i, v in enumerate(Meas):
            if frontend.data is None:
                Table[self.corpus_pos, 1:] = np.nan
                break
            value = getattr(frontend, v)()
            value = value if value is not None else np.nan
            Table[self.corpus_pos, i+1] = value

        if hasattr(frontend, '_check'):
            frontend._check()


        if self.is_last_expe():
            tablefmt = 'latex' if fmt == 'tex' else fmt
            precision = '.5f'
            print(colored('\nStats Table :', 'green'))
            Meas_ = self.D.Meas_
            print(self.tabulate(Table, headers=Meas_, tablefmt=tablefmt, floatfmt=precision))


